import factory

from spectator.core.factories import IndividualCreatorFactory

from . import models


class VenueFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Venue

    name = factory.Sequence(lambda n: f"Venue {n}")


class WorkFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Work

    title = factory.Sequence(lambda n: f"Work {n}")


class ClassicalWorkFactory(WorkFactory):
    kind = "classicalwork"
    title = factory.Sequence(lambda n: f"Classical Work {n}")


class DancePieceFactory(WorkFactory):
    kind = "dancepiece"
    title = factory.Sequence(lambda n: f"Dance Piece {n}")


class ExhibitionFactory(WorkFactory):
    kind = "exhibition"
    title = factory.Sequence(lambda n: f"Exhibition {n}")


class MovieFactory(WorkFactory):
    kind = "movie"
    title = factory.Sequence(lambda n: f"Movie {n}")


class PlayFactory(WorkFactory):
    kind = "play"
    title = factory.Sequence(lambda n: f"Play {n}")


class WorkRoleFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.WorkRole

    role_name = factory.Sequence(lambda n: f"Role {n}")
    creator = factory.SubFactory(IndividualCreatorFactory)
    work = factory.SubFactory(WorkFactory)


class EventFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Event

    title = factory.Sequence(lambda n: f"Event {n}")
    venue = factory.SubFactory(VenueFactory)
    # Bigger width/height than default detail_thumbnail_2x size:
    thumbnail = factory.django.ImageField(color="blue", width=800, height=800)


class ComedyEventFactory(EventFactory):
    kind = "comedy"


class ConcertEventFactory(EventFactory):
    kind = "concert"


class DanceEventFactory(EventFactory):
    kind = "dance"


class MuseumEventFactory(EventFactory):
    kind = "museum"


class GigEventFactory(EventFactory):
    kind = "gig"


class MiscEventFactory(EventFactory):
    kind = "misc"


class CinemaEventFactory(EventFactory):
    kind = "cinema"


class TheatreEventFactory(EventFactory):
    kind = "theatre"


class EventRoleFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.EventRole

    role_name = factory.Sequence(lambda n: f"Role {n}")
    creator = factory.SubFactory(IndividualCreatorFactory)
    event = factory.SubFactory(MiscEventFactory)


class WorkSelectionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.WorkSelection

    event = factory.SubFactory(MiscEventFactory)
    work = factory.SubFactory(WorkFactory)



# Generated by Django 2.0 on 2018-01-02 09:59

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0004_venue_slug"),
    ]

    operations = [
        migrations.AlterField(
            model_name="event",
            name="movie",
            field=models.ForeignKey(
                blank=True,
                help_text="Only used if event is of 'Movie' kind.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="spectator_events.Movie",
            ),
        ),
        migrations.AlterField(
            model_name="event",
            name="play",
            field=models.ForeignKey(
                blank=True,
                help_text="Only used if event is of 'Play' kind.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="spectator_events.Play",
            ),
        ),
    ]



import csv

from django.core.management.base import BaseCommand, CommandError

from spectator.events.models import Event


class Command(BaseCommand):
    """
    Generates a CSV file containing information about Works of kind
    "movie" that have been seen, suitable for importing into a
    Letterboxd.com account.

    Import docs: https://letterboxd.com/about/importing-data/
    """

    help = (
        "Generates a CSV file of watched movies suitable for import into Letterboxd.com"
    )

    filename = "watched_movies.csv"

    def handle(self, *args, **options):
        "This is called when the command is run."
        rows = self.make_rows()

        if len(rows) == 0:
            msg = "No movies were found."
            raise CommandError(msg)

        self.write_csv_file(rows)

        plural = "movie" if len(rows) == 1 else "movies"
        self.stdout.write(
            self.style.SUCCESS(f"Wrote {len(rows)} {plural} to {self.filename}")
        )

    def make_rows(self):
        """
        Returns a list of dicts, each one about a single viewing of
        a movie.
        """
        rows = []
        watched_work_ids = []

        for event in Event.objects.filter(kind="cinema").order_by("date"):
            for selection in event.get_movies():
                work = selection.work
                is_rewatch = work.id in watched_work_ids

                directors = []
                for role in work.roles.all():
                    if role.role_name.lower() == "director":
                        directors.append(role.creator.name)

                rows.append(
                    {
                        "imdbID": work.imdb_id,
                        "Title": work.title,
                        "Year": work.year,
                        "Directors": ", ".join(directors),
                        "WatchedDate": event.date.strftime("%Y-%m-%d"),
                        "Rewatch": str(is_rewatch).lower(),
                    }
                )

                watched_work_ids.append(work.id)

        return rows

    def write_csv_file(self, rows):
        """
        Passed a list of dicts - each one being data about single
        viewing of a movie - writes this out to a CSV file.
        """
        with open(self.filename, mode="w") as movies_file:
            writer = csv.DictWriter(
                movies_file,
                fieldnames=rows[0].keys(),
                delimiter=",",
                quotechar='"',
                quoting=csv.QUOTE_MINIMAL,
            )

            writer.writeheader()

            for row in rows:
                writer.writerow(row)



# Generated by Django 2.0 on 2018-02-08 14:17

from django.conf import settings
from django.db import migrations
from hashids import Hashids


def generate_slug(value):
    """
    Generates a slug using a Hashid of `value`.

    Taken from spectator_core.models.SluggedModelMixin
    """
    # Defaults:
    alphabet = "abcdefghijkmnopqrstuvwxyz23456789"
    salt = "Django Spectator"

    if hasattr(settings, "SPECTATOR_SLUG_ALPHABET"):
        alphabet = settings.SPECTATOR_SLUG_ALPHABET

    if hasattr(settings, "SPECTATOR_SLUG_SALT"):
        salt = settings.SPECTATOR_SLUG_SALT

    hashids = Hashids(alphabet=alphabet, salt=salt, min_length=5)

    return hashids.encode(value)


def forwards(apps, schema_editor):
    """
    Re-save all the Works because something earlier didn't create their slugs.
    """
    Work = apps.get_model("spectator_events", "Work")

    for work in Work.objects.all():
        if not work.slug:
            work.slug = generate_slug(work.pk)
            work.save()


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0031_auto_20180208_1412"),
    ]

    operations = [
        migrations.RunPython(forwards),
    ]



from django.db.models import Min
from django.http import Http404
from django.utils.encoding import force_str
from django.utils.translation import gettext as _
from django.views.generic import DetailView, ListView, YearArchiveView
from django.views.generic.detail import SingleObjectMixin

from spectator.core.views import PaginatedListView

from .models import Publication, PublicationSeries, Reading


class ReadingHomeView(ListView):
    model = Publication
    template_name = "spectator_reading/home.html"
    queryset = (
        Publication.unread_objects.select_related("series")
        .prefetch_related("roles__creator")
        .all()
    )
    ordering = ["time_created"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["in_progress_publication_list"] = (
            Publication.in_progress_objects.select_related("series")
            .prefetch_related("roles__creator")
            .all()
        )
        return context


class PublicationSeriesListView(ListView):
    model = PublicationSeries


class PublicationSeriesDetailView(SingleObjectMixin, PaginatedListView):
    template_name = "spectator_reading/publicationseries_detail.html"

    def get(self, request, *args, **kwargs):
        self.object = self.get_object(queryset=PublicationSeries.objects.all())
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["publicationseries"] = self.object
        context["publication_list"] = context["object_list"]
        return context

    def get_queryset(self):
        return (
            self.object.publication_set.select_related("series")
            .prefetch_related("roles__creator")
            .all()
        )


class PublicationListView(PaginatedListView):
    model = Publication
    publication_kind = "book"

    def get(self, request, *args, **kwargs):
        # Are we should 'book's (default) or 'periodical's?
        if self.kwargs.get("kind", None) == "periodical":
            self.publication_kind = "periodical"
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["publication_kind"] = self.publication_kind
        context["book_count"] = Publication.objects.filter(kind="book").count()
        context["periodical_count"] = Publication.objects.filter(
            kind="periodical"
        ).count()
        return context

    def get_queryset(self):
        qs = super().get_queryset()
        qs = (
            qs.filter(kind=self.publication_kind)
            .select_related("series")
            .prefetch_related("roles__creator")
        )
        return qs

    def get_ordering(self):
        if self.publication_kind == "periodical":
            return ("series__title_sort", "title_sort")
        else:
            return ("title_sort",)


class PublicationDetailView(DetailView):
    model = Publication


class ReadingYearArchiveView(YearArchiveView):
    allow_empty = True
    date_field = "end_date"
    make_object_list = True
    model = Reading
    ordering = "end_date"
    # Could be set to 'periodical' or 'book' in get():
    publication_kind = None
    # Will be a QS of all publications finished this year:
    all_publications_queryset = None

    def get(self, request, *args, **kwargs):
        # Are we should 'book's (default) or 'periodical's?
        kind = self.kwargs.get("kind", None)
        if kind == "periodicals":
            self.publication_kind = "periodical"
        elif kind == "books":
            self.publication_kind = "book"
        elif kind is not None:
            msg = f"'{kind}' is not a valid publication kind"
            raise Http404(msg)

        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["publication_kind"] = self.publication_kind

        context["publication_count"] = self.all_publications_queryset.count()
        context["book_count"] = self.all_publications_queryset.filter(
            publication__kind="book"
        ).count()
        context["periodical_count"] = self.all_publications_queryset.filter(
            publication__kind="periodical"
        ).count()

        # Should we divide the readings up by month in the template?
        context["show_months"] = True

        objects = context["object_list"]
        if (
            len(objects) > 1
            and objects.first().end_granularity == Reading.DateGranularity.YEAR
            and objects.last().end_granularity == Reading.DateGranularity.YEAR
        ):
            # If the first and last publications only have year-based
            # granularity, assume everything does, so we shouldn't
            # show the months in the template.
            context["show_months"] = False

        return context

    def get_queryset(self):
        "Reduce the number of queries and speed things up."
        qs = super().get_queryset()

        qs = qs.select_related("publication__series").prefetch_related(
            "publication__roles__creator"
        )

        return qs

    def get_dated_items(self):
        items, qs, info = super().get_dated_items()

        if "year" in info and info["year"]:
            # Get the earliest date we have a Reading for:
            end_date_min = Reading.objects.aggregate(Min("end_date"))["end_date__min"]
            # Make it a 'yyyy-01-01' date:
            min_year_date = end_date_min.replace(month=1, day=1)
            if info["year"] < min_year_date:
                # The year we're viewing is before our minimum date, so 404.
                raise Http404(
                    _("No %(verbose_name_plural)s available")
                    % {
                        "verbose_name_plural": force_str(
                            qs.model._meta.verbose_name_plural
                        )
                    }
                )
            elif info["year"] == min_year_date:
                # This is the earliest year we have readings for, so
                # there is no previous year.
                info["previous_year"] = None

        # Save the QuerySet for ALL kinds for use in get_context_data():
        self.all_publications_queryset = qs

        # Now filter the results if necessary:
        if self.publication_kind is not None:
            qs = qs.filter(publication__kind=self.publication_kind)

        return items, qs, info



# Generated by Django 1.11 on 2017-11-01 09:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_core", "0002_creator_slug"),
    ]

    operations = [
        migrations.AlterField(
            model_name="creator",
            name="slug",
            field=models.SlugField(blank=True, max_length=10),
        ),
    ]



# Generated by Django 2.0 on 2018-01-18 09:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0009_event_note"),
    ]

    operations = [
        migrations.AlterField(
            model_name="event",
            name="note",
            field=models.TextField(
                blank=True,
                help_text="Optional. Paragraphs will be surrounded with &lt;p&gt;&lt;/p&gt; tags. HTML allowed.",  # noqa: E501
            ),
        ),
    ]



# Generated by Django 5.1.4 on 2025-04-22 08:56

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spectator_events', '0045_auto_20201221_1358'),
    ]

    operations = [
        migrations.AlterField(
            model_name='work',
            name='imdb_id',
            field=models.CharField(blank=True, help_text='Starts with \'tt\', e.g. \'tt0100842\'.\n            From <a href="https://www.imdb.com">IMDb</a>.', max_length=12, validators=[django.core.validators.RegexValidator(code='invalid_imdb_id', message='IMDb ID should be like "tt1234567"', regex='^tt[0-9]{7,10}$')], verbose_name='IMDb ID'),
        ),
    ]



# Generated by Django 2.0 on 2018-02-14 13:24

from django.db import migrations


def forwards(apps, schema_editor):
    """
    Change Events with kind 'movie' to 'cinema'
    and Events with kind 'play' to 'theatre'.

    Purely for more consistency.
    """
    Event = apps.get_model("spectator_events", "Event")

    for ev in Event.objects.filter(kind="movie"):
        ev.kind = "cinema"
        ev.save()

    for ev in Event.objects.filter(kind="play"):
        ev.kind = "theatre"
        ev.save()


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0034_auto_20180208_1618"),
    ]

    operations = [
        migrations.RunPython(forwards),
    ]



from django.utils.html import strip_tags
from django.utils.text import Truncator


def truncate_string(
    text, *, strip_html=True, chars=255, truncate="…", at_word_boundary=False
):
    """Truncate a string to a certain length, removing line breaks and mutliple
    spaces, optionally removing HTML, and appending a 'truncate' string.

    Keyword arguments:
    strip_html -- boolean.
    chars -- Number of characters to return.
    at_word_boundary -- Only truncate at a word boundary, which will probably
        result in a string shorter than chars.
    truncate -- String to add to the end.
    """
    if strip_html:
        text = strip_tags(text)
    text = text.replace("\n", " ").replace("\r", "")
    text = " ".join(text.split())
    if at_word_boundary:
        if len(text) > chars:
            text = text[:chars].rsplit(" ", 1)[0] + truncate
    else:
        text = Truncator(text).chars(chars, html=False, truncate=truncate)
    return text


def chartify(qs, score_field, *, cutoff=0, ensure_chartiness=True):
    """
    Given a QuerySet it will go through and add a `chart_position` property to
    each object returning a list of the objects.

    If adjacent objects have the same 'score' (based on `score_field`) then
    they will have the same `chart_position`. This can then be used in
    templates for the `value` of <li> elements in an <ol>.

    By default any objects with a score of 0 or less will be removed.

    By default, if all the items in the chart have the same position, no items
    will be returned (it's not much of a chart).

    Keyword arguments:
    qs -- The QuerySet
    score_field -- The name of the numeric field that each object in the
                   QuerySet has, that will be used to compare their positions.
    cutoff -- Any objects with a score of this value or below will be removed
              from the list. Set to None to disable this.
    ensure_chartiness -- If True, then if all items in the list have the same
                         score, an empty list will be returned.
    """
    chart = []
    position = 0
    prev_obj = None

    for counter, obj in enumerate(qs):
        score = getattr(obj, score_field)

        if score != getattr(prev_obj, score_field, None):
            position = counter + 1

        if cutoff is None or score > cutoff:
            obj.chart_position = position
            chart.append(obj)

        prev_obj = obj

    if (
        ensure_chartiness
        and len(chart) > 0
        and getattr(chart[0], score_field) == getattr(chart[-1], score_field)
    ):
        chart = []

    return chart



# Generated by Django 2.0 on 2018-02-08 14:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0030_movies_to_works"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="event",
            name="classicalworks",
        ),
        migrations.RemoveField(
            model_name="event",
            name="dancepieces",
        ),
        migrations.RemoveField(
            model_name="event",
            name="movies",
        ),
        migrations.RemoveField(
            model_name="event",
            name="plays",
        ),
        migrations.AddField(
            model_name="event",
            name="works",
            field=models.ManyToManyField(
                blank=True,
                through="spectator_events.WorkSelection",
                to="spectator_events.Work",
            ),
        ),
    ]



# Generated by Django 2.0 on 2018-01-02 11:27

from django.conf import settings
from django.db import migrations, models
from hashids import Hashids


def generate_slug(value):
    "A copy of spectator.core.models.SluggedModelMixin._generate_slug()"
    alphabet = "abcdefghijkmnopqrstuvwxyz23456789"
    salt = "Django Spectator"

    if hasattr(settings, "SPECTATOR_SLUG_ALPHABET"):
        alphabet = settings.SPECTATOR_SLUG_ALPHABET

    if hasattr(settings, "SPECTATOR_SLUG_SALT"):
        salt = settings.SPECTATOR_SLUG_SALT

    hashids = Hashids(alphabet=alphabet, salt=salt, min_length=5)

    return hashids.encode(value)


def set_slug(apps, schema_editor):
    """
    Create a slug for each Event already in the DB.
    """
    Event = apps.get_model("spectator_events", "Event")

    for e in Event.objects.all():
        e.slug = generate_slug(e.pk)
        e.save(update_fields=["slug"])


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0005_auto_20180102_0959"),
    ]

    operations = [
        migrations.AlterField(
            model_name="event",
            name="slug",
            field=models.SlugField(blank=True, default="a", max_length=10),
            preserve_default=False,
        ),
        migrations.RunPython(set_slug),
    ]



# Generated by Django 2.0 on 2018-02-08 16:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0032_recreate_work_slugs"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="classicalwork",
            name="creators",
        ),
        migrations.RemoveField(
            model_name="classicalworkrole",
            name="classical_work",
        ),
        migrations.RemoveField(
            model_name="classicalworkrole",
            name="creator",
        ),
        migrations.RemoveField(
            model_name="classicalworkselection",
            name="event",
        ),
        migrations.RemoveField(
            model_name="classicalworkselection",
            name="work",
        ),
        migrations.RemoveField(
            model_name="dancepiece",
            name="creators",
        ),
        migrations.RemoveField(
            model_name="dancepiecerole",
            name="creator",
        ),
        migrations.RemoveField(
            model_name="dancepiecerole",
            name="dance_piece",
        ),
        migrations.RemoveField(
            model_name="dancepieceselection",
            name="event",
        ),
        migrations.RemoveField(
            model_name="dancepieceselection",
            name="work",
        ),
        migrations.RemoveField(
            model_name="movie",
            name="creators",
        ),
        migrations.RemoveField(
            model_name="movierole",
            name="creator",
        ),
        migrations.RemoveField(
            model_name="movierole",
            name="movie",
        ),
        migrations.RemoveField(
            model_name="movieselection",
            name="event",
        ),
        migrations.RemoveField(
            model_name="movieselection",
            name="work",
        ),
        migrations.RemoveField(
            model_name="play",
            name="creators",
        ),
        migrations.RemoveField(
            model_name="playrole",
            name="creator",
        ),
        migrations.RemoveField(
            model_name="playrole",
            name="play",
        ),
        migrations.RemoveField(
            model_name="playselection",
            name="event",
        ),
        migrations.RemoveField(
            model_name="playselection",
            name="work",
        ),
        migrations.AlterField(
            model_name="event",
            name="kind",
            field=models.CharField(
                choices=[
                    ("concert", "Classical"),
                    ("comedy", "Comedy"),
                    ("dance", "Dance"),
                    ("exhibition", "Exhibition"),
                    ("gig", "Gig"),
                    ("misc", "Other"),
                    ("movie", "Movie"),
                    ("play", "Theatre"),
                ],
                max_length=20,
            ),
        ),
        migrations.DeleteModel(
            name="ClassicalWork",
        ),
        migrations.DeleteModel(
            name="ClassicalWorkRole",
        ),
        migrations.DeleteModel(
            name="ClassicalWorkSelection",
        ),
        migrations.DeleteModel(
            name="DancePiece",
        ),
        migrations.DeleteModel(
            name="DancePieceRole",
        ),
        migrations.DeleteModel(
            name="DancePieceSelection",
        ),
        migrations.DeleteModel(
            name="Movie",
        ),
        migrations.DeleteModel(
            name="MovieRole",
        ),
        migrations.DeleteModel(
            name="MovieSelection",
        ),
        migrations.DeleteModel(
            name="Play",
        ),
        migrations.DeleteModel(
            name="PlayRole",
        ),
        migrations.DeleteModel(
            name="PlaySelection",
        ),
    ]



from imagekit import ImageSpec, register
from imagekit.processors import ResizeToFit

from spectator.core import app_settings

# NOTE: All of these generators are deprecated.
# Use the thumbnail properties on Publication and Reading models instead.


class Thumbnail(ImageSpec):
    "Base class"

    format = "JPEG"
    options = {"quality": 80}


class ListThumbnail(Thumbnail):
    "For displaying in lists of Publications, Events, etc."

    processors = [ResizeToFit(*app_settings.THUMBNAIL_LIST_SIZE)]


class ListThumbnail2x(ListThumbnail):
    """Retina version of ListThumbnail
    Generated twice the size of our set dimensions.
    """

    dimensions = [d * 2 for d in app_settings.THUMBNAIL_LIST_SIZE]
    processors = [ResizeToFit(*dimensions)]


class DetailThumbnail(Thumbnail):
    "For displaying on the detail pages of Publication, Event, etc"

    processors = [ResizeToFit(*app_settings.THUMBNAIL_DETAIL_SIZE)]


class DetailThumbnail2x(DetailThumbnail):
    """Retina version of DetailThumbnail
    Generated twice the size of our set dimensions.
    """

    dimensions = [d * 2 for d in app_settings.THUMBNAIL_DETAIL_SIZE]
    processors = [ResizeToFit(*dimensions)]


register.generator("spectator:list_thumbnail", ListThumbnail)
register.generator("spectator:list_thumbnail2x", ListThumbnail2x)
register.generator("spectator:detail_thumbnail", DetailThumbnail)
register.generator("spectator:detail_thumbnail2x", DetailThumbnail2x)



from django.contrib.sitemaps import Sitemap

from .models import Creator


class CreatorSitemap(Sitemap):
    changefreq = "yearly"
    priority = 0.5

    def items(self):
        return Creator.objects.all()

    def lastmod(self, obj):
        return obj.time_modified



# Generated by Django 1.11 on 2017-04-17 17:14

from django.db import migrations, models

import spectator.core.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Creator",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The time this item was created in the database.",
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The time this item was last saved to the database.",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="e.g. 'Douglas Adams' or 'The Long Blondes'.",
                        max_length=255,
                    ),
                ),
                (
                    "name_sort",
                    spectator.core.fields.NaturalSortField(
                        "name",
                        db_index=True,
                        default="",
                        editable=False,
                        help_text="Best for sorting groups. e.g. 'long blondes, the' or 'adams, douglas'.",  # noqa: E501
                        max_length=255,
                    ),
                ),
                (
                    "kind",
                    models.CharField(
                        choices=[("individual", "Individual"), ("group", "Group")],
                        default="individual",
                        max_length=20,
                    ),
                ),
            ],
            options={"ordering": ("name_sort",)},
        ),
    ]



# Generated by Django 2.0 on 2018-04-17 12:44

from django.db import migrations
from hashids import Hashids

from spectator.core import app_settings


def generate_slug(value):
    """
    Generates a slug using a Hashid of `value`.

    COPIED from spectator.core.models.SluggedModelMixin() because migrations
    don't make this happen automatically and perhaps the least bad thing is
    to copy the method here, ugh.
    """
    alphabet = app_settings.SLUG_ALPHABET
    salt = app_settings.SLUG_SALT

    hashids = Hashids(alphabet=alphabet, salt=salt, min_length=5)

    return hashids.encode(value)


def forwards(apps, schema_editor):
    """
    Having added the new 'exhibition' Work type, we're going to assume that
    every Event of type 'museum' should actually have one Exhibition attached.

    So, we'll add one, with the same title as the Event.
    And we'll move all Creators from the Event to the Exhibition.
    """
    Event = apps.get_model("spectator_events", "Event")
    Work = apps.get_model("spectator_events", "Work")
    WorkRole = apps.get_model("spectator_events", "WorkRole")
    WorkSelection = apps.get_model("spectator_events", "WorkSelection")

    for event in Event.objects.filter(kind="museum"):

        # Create a new Work based on this Event's details.

        work = Work.objects.create(
            kind="exhibition", title=event.title, title_sort=event.title_sort
        )
        # This doesn't generate the slug field automatically because Django.
        # So we'll have to do it manually. Graarhhh.
        work.slug = generate_slug(work.pk)
        work.save()

        # Associate the new Work with the Event.
        WorkSelection.objects.create(event=event, work=work)

        # Associate any Creators on the Event with the new Work.
        for role in event.roles.all():
            WorkRole.objects.create(
                creator=role.creator,
                work=work,
                role_name=role.role_name,
                role_order=role.role_order,
            )

            # Remove Creators from the Event.
            role.delete()


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0038_auto_20180417_1224"),
    ]

    operations = [
        migrations.RunPython(forwards),
    ]



from collections import OrderedDict

from django.db.models import Count
from django.db.models.functions import TruncYear

from .models import Reading


def annual_reading_counts(kind="all"):
    """
    Returns a list of dicts, one per year of reading. In year order.
    Each dict is like this (if kind is 'all'):

        {'year':        datetime.date(2003, 1, 1),
         'book':        12,    # only included if kind is 'all' or 'book'
         'periodical':  18,    # only included if kind is 'all' or 'periodical'
         'total':       30,    # only included if kind is 'all'
         }

    We use the end_date of a Reading to count when that thing was read.

    kind is one of 'book', 'periodical' or 'all', for both.
    """
    kinds = ["book", "periodical"] if kind == "all" else [kind]

    # This will have keys of years (strings) and dicts of data:
    # {
    #   '2003': {'books': 12, 'periodicals': 18},
    # }
    counts = OrderedDict()

    for k in kinds:
        qs = (
            Reading.objects.exclude(end_date__isnull=True)
            .filter(publication__kind=k)
            .annotate(year=TruncYear("end_date"))
            .values("year")
            .annotate(count=Count("id"))
            .order_by("year")
        )

        for year_data in qs:
            year_str = year_data["year"].strftime("%Y")
            if year_str not in counts:
                counts[year_str] = {
                    "year": year_data["year"],
                }

            counts[year_str][k] = year_data["count"]

    # Now translate counts into our final list, with totals, and 0s for kinds
    # when they have no Readings for that year.
    counts_list = []

    for _year_str, data in counts.items():
        year_data = {
            "year": data["year"],
        }
        if kind == "all":
            year_data["total"] = 0

        for k in kinds:
            if k in data:
                year_data[k] = data[k]
                if kind == "all":
                    year_data["total"] += data[k]
            else:
                year_data[k] = 0

        counts_list.append(year_data)

    return counts_list



from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse

from spectator.core.fields import NaturalSortField
from spectator.core.models import (
    BaseRole,
    SluggedModelMixin,
    ThumbnailModelMixin,
    TimeStampedModelMixin,
    thumbnail_upload_path,
)

from . import managers


def publication_upload_path(instance, filename):
    """
    This function is now only kept so that older migrations still work.

    No longer needed since moving the Publication.thumbnail field to the
    ThumbnailModelMixin.
    2020-04-07
    """
    return thumbnail_upload_path(instance, filename)


class PublicationSeries(TimeStampedModelMixin, SluggedModelMixin, models.Model):
    """
    A way to group `Publication`s into series.

    Get its Publications:

        series.publication_set.all()
    """

    title = models.CharField(
        null=False,
        blank=False,
        max_length=255,
        help_text="e.g. 'The London Review of Books'.",
    )

    title_sort = NaturalSortField(
        "title",
        max_length=255,
        default="",
        help_text="e.g. 'london review of books, the'.",
    )

    url = models.URLField(
        null=False,
        blank=True,
        max_length=255,
        verbose_name="URL",
        help_text="e.g. 'https://www.lrb.co.uk/'.",
    )

    class Meta:
        ordering = ("title_sort",)
        verbose_name = "Publication series"
        verbose_name_plural = "Publication series"

    def __str__(self):
        return self.title

    def get_absolute_url(self):
        return reverse(
            "spectator:reading:publicationseries_detail", kwargs={"slug": self.slug}
        )


class PublicationRole(BaseRole):
    """
    Linking a creator to a Publication, optionally via their role (e.g.
    'Author', 'Editor', etc.)
    """

    creator = models.ForeignKey(
        "spectator_core.Creator",
        blank=False,
        on_delete=models.CASCADE,
        related_name="publication_roles",
    )

    publication = models.ForeignKey(
        "spectator_reading.Publication", on_delete=models.CASCADE, related_name="roles"
    )

    class Meta:
        ordering = ("role_order", "role_name")
        verbose_name = "Publication role"


class Publication(ThumbnailModelMixin, TimeStampedModelMixin, SluggedModelMixin):
    """
    Get a Publication's creators:

        publication = Publication.objects.get(pk=1)

        # Just the creators:
        for creator in publication.creators.all():
            print(creator.name)

        # Include their roles:
        for role in publication.roles.all():
            print(role.publication, role.creator, role.role_name)

    Get its readings:

        for reading in publication.reading_set.all():
            print(reading.start_date, reading.end_date)
    """

    class Kind(models.TextChoices):
        BOOK = "book", "Book"
        PERIODICAL = "periodical", "Periodical"

    title = models.CharField(
        null=False,
        blank=False,
        max_length=255,
        help_text="e.g. 'Aurora' or 'Vol. 39 No. 4, 16 February 2017'.",
    )

    title_sort = NaturalSortField(
        "title",
        max_length=255,
        default="",
        help_text="e.g. 'clockwork orange, a' or 'world cities, the'.",
    )

    series = models.ForeignKey(
        "spectator_reading.PublicationSeries",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )

    kind = models.CharField(max_length=20, choices=Kind.choices, default=Kind.BOOK)

    official_url = models.URLField(
        null=False,
        blank=True,
        max_length=255,
        verbose_name="Official URL",
        help_text="Official URL for this book/issue.",
    )

    isbn_uk = models.CharField(
        null=False,
        blank=True,
        max_length=20,
        verbose_name="UK ISBN",
        help_text="e.g. '0356500489'.",
    )

    isbn_us = models.CharField(
        null=False,
        blank=True,
        max_length=20,
        verbose_name="US ISBN",
        help_text="e.g. '0316098094'.",
    )

    notes_url = models.URLField(
        null=False,
        blank=True,
        max_length=255,
        verbose_name="Notes URL",
        help_text="URL of your notes/review.",
    )

    creators = models.ManyToManyField(
        "spectator_core.Creator", through="PublicationRole", related_name="publications"
    )

    # Managers

    objects = models.Manager()

    # Publications that are currently being read:
    in_progress_objects = managers.InProgressPublicationsManager()

    # Publications that haven't been started (have no Readings):
    unread_objects = managers.UnreadPublicationsManager()

    class Meta:
        ordering = ("title_sort",)

    def __str__(self):
        return self.title

    def get_absolute_url(self):
        return reverse(
            "spectator:reading:publication_detail", kwargs={"slug": self.slug}
        )

    def get_current_reading(self):
        try:
            return self.reading_set.filter(end_date__isnull=True)[0]
        except IndexError:
            pass

    @property
    def amazon_uk_url(self):
        url = ""
        if self.isbn_uk:
            url = f"https://www.amazon.co.uk/gp/product/{self.isbn_uk}/"
            if (
                hasattr(settings, "SPECTATOR_AMAZON")
                and "uk" in settings.SPECTATOR_AMAZON
            ):
                url = "{}?tag={}".format(url, settings.SPECTATOR_AMAZON["uk"])
        return url

    @property
    def amazon_us_url(self):
        url = ""
        if self.isbn_us:
            url = f"https://www.amazon.com/dp/{self.isbn_us}/"
            if (
                hasattr(settings, "SPECTATOR_AMAZON")
                and "us" in settings.SPECTATOR_AMAZON
            ):
                url = "{}?tag={}".format(url, settings.SPECTATOR_AMAZON["us"])
        return url

    @property
    def amazon_urls(self):
        urls = []
        if self.isbn_uk:
            urls.append(
                {"url": self.amazon_uk_url, "name": "Amazon.co.uk", "country": "UK"}
            )
        if self.isbn_us:
            urls.append(
                {"url": self.amazon_us_url, "name": "Amazon.com", "country": "USA"}
            )
        return urls

    @property
    def has_urls(self):
        "Handy for templates."
        if self.isbn_uk or self.isbn_us or self.official_url or self.notes_url:  # noqa: SIM103
            return True
        else:
            return False


class Reading(TimeStampedModelMixin, models.Model):
    """
    A period when a Publication was read.
    """

    class DateGranularity(models.IntegerChoices):
        # Via https://www.flickr.com/services/api/misc.dates.html
        # SECOND = 0, "Y-m-d H:i:s"
        DAY = 3, "Y-m-d"
        MONTH = 4, "Y-m"
        YEAR = 6, "Y"
        # CIRCA = 8, "Circa..."

    publication = models.ForeignKey(
        "spectator_reading.Publication",
        null=False,
        blank=False,
        on_delete=models.CASCADE,
    )

    start_date = models.DateField(null=True, blank=True)
    start_granularity = models.PositiveSmallIntegerField(
        null=False,
        blank=False,
        default=DateGranularity.DAY,
        choices=DateGranularity.choices,
    )
    end_date = models.DateField(null=True, blank=True)
    end_granularity = models.PositiveSmallIntegerField(
        null=False,
        blank=False,
        default=DateGranularity.DAY,
        choices=DateGranularity.choices,
    )
    is_finished = models.BooleanField(
        default=False, help_text="Did you finish the publication?"
    )

    objects = managers.EndDateAscendingReadingsManager()
    objects_desc = managers.EndDateDescendingReadingsManager()

    def __str__(self):
        return f"{self.publication} ({self.start_date} to {self.end_date})"

    def clean(self):
        if self.start_date and self.end_date and self.start_date > self.end_date:
            msg = "A Reading's end date can't be before its start date."
            raise ValidationError(msg)



from django.contrib import admin
from django.db.models import Count
from django.templatetags.l10n import unlocalize
from imagekit.admin import AdminThumbnail

from spectator.core import app_settings

from .models import Event, EventRole, Venue, Work, WorkRole, WorkSelection

# INLINES


class EventRoleInline(admin.TabularInline):
    model = EventRole
    fields = ("creator", "role_name", "role_order")
    raw_id_fields = ("creator",)
    extra = 0


class WorkRoleInline(admin.TabularInline):
    model = WorkRole
    fields = ("creator", "role_name", "role_order")
    raw_id_fields = ("creator",)
    extra = 0


class WorkSelectionInline(admin.TabularInline):
    model = WorkSelection
    fields = ("work", "order")
    raw_id_fields = ("work",)
    extra = 0


# MODEL ADMINS.


@admin.register(Event)
class EventAdmin(admin.ModelAdmin):
    list_display = ("__str__", "date", "list_thumbnail", "kind_name", "venue")
    list_filter = ("kind", "date")
    search_fields = ("title",)

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "kind",
                    "date",
                    "venue",
                    "venue_name",
                    "title",
                    "title_sort",
                    "detail_thumbnail",
                    "thumbnail",
                    "slug",
                    "note",
                )
            },
        ),
        (
            "Times",
            {"classes": ("collapse",), "fields": ("time_created", "time_modified")},
        ),
    )

    raw_id_fields = ("venue",)
    readonly_fields = (
        "title_sort",
        "slug",
        "detail_thumbnail",
        "time_created",
        "time_modified",
    )

    inlines = [WorkSelectionInline, EventRoleInline]

    detail_thumbnail = AdminThumbnail(
        image_field="thumbnail", template="spectator_core/admin/detail_thumbnail.html"
    )

    list_thumbnail = AdminThumbnail(
        image_field="thumbnail", template="spectator_core/admin/list_thumbnail.html"
    )

    def save_related(self, request, form, formsets, change):
        """
        When the Admin saves the Event, the related M2M things, like Movies,
        haven't yet been saved. So when the Event's title_sort field has
        nothing to go on.

        So, cheekily, after the parent's save_related() method has saved
        all the M2M fields, we save the Event again. This time its
        title_sort will be set correctly.
        """
        super().save_related(request, form, formsets, change)
        form.instance.save()


@admin.register(Work)
class WorkAdmin(admin.ModelAdmin):
    list_display = ("title", "kind", "tidy_year")
    search_fields = ("title",)
    list_filter = ("kind", "year")

    fieldsets = (
        (None, {"fields": ("kind", "title", "title_sort", "slug", "year", "imdb_id")}),
        (
            "Times",
            {"classes": ("collapse",), "fields": ("time_created", "time_modified")},
        ),
    )

    readonly_fields = ("title_sort", "slug", "time_created", "time_modified")
    inlines = [WorkRoleInline]

    def tidy_year(self, obj):
        "Stop the year appearing like '2,018' when USE_THOUSAND_SEPARATOR=True"
        if obj.year:
            return unlocalize(obj.year)
        else:
            return "-"

    tidy_year.short_description = "Year"


class CountryListFilter(admin.SimpleListFilter):
    """
    For filtering Venues by country.

    Used so that we only list countries that have at least one Venue assigned
    to them.

    Otherwise we have a very long list of countries, most of which are unused.
    """

    title = "Country"

    parameter_name = "country"

    def lookups(self, request, model_admin):
        """
        Returns a list of tuples like:

            [
                ('AU', 'Australia'),
                ('GB', 'UK'),
                ('US', 'USA'),
            ]

        One for each country that has at least one Venue.
        Sorted by the label names.
        """
        list_of_countries = []

        # We don't need the country_count but we need to annotate them in order
        # to group the results.
        qs = (
            Venue.objects.exclude(country="")
            .values("country")
            .annotate(country_count=Count("country"))
            .order_by("country")
        )
        for obj in qs:
            country = obj["country"]
            list_of_countries.append((country, Venue.COUNTRIES[country]))

        return sorted(list_of_countries, key=lambda c: c[1])

    def queryset(self, request, queryset):
        """
        Returns the filtered queryset based on the value
        provided in the query string and retrievable via
        `self.value()`.
        """
        if self.value():
            return queryset.filter(country=self.value())
        else:
            return queryset


@admin.register(Venue)
class VenueAdmin(admin.ModelAdmin):
    list_display = ("name", "address", "country")
    list_filter = (CountryListFilter,)
    search_fields = ("name",)

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "name_sort",
                    "note",
                    "cinema_treasures_id",
                    "latitude",
                    "longitude",
                    "address",
                    "country",
                )
            },
        ),
        (
            "Times",
            {"classes": ("collapse",), "fields": ("time_created", "time_modified")},
        ),
    )

    readonly_fields = ("name_sort", "time_created", "time_modified")

    class Media:
        """
        If maps are enabled then we add the JS and CSS for either
        Google JS Maps API or the Mapbox APIs (including Geocoding).
        """

        if app_settings.MAPS["enable"] is True:
            # Add the CSS and JS for Google or Mapbox maps and geocoding.
            library_css = ()
            library_js = ()

            if app_settings.MAPS["library"] == "google":
                library_js = (
                    "https://maps.googleapis.com/maps/api/js?key={}".format(
                        app_settings.MAPS["api_key"]
                    ),
                    "js/admin/location_picker_google.js",
                )

            elif app_settings.MAPS["library"] == "mapbox":
                library_css = (
                    "https://api.tiles.mapbox.com/mapbox-gl-js/v1.6.1/mapbox-gl.css",
                )
                library_js = (
                    "https://api.tiles.mapbox.com/mapbox-gl-js/v1.6.1/mapbox-gl.js",
                    "js/admin/location_picker_mapbox.js",
                )

            css = {"all": library_css + ("css/admin/location_picker.css",)}
            js = library_js

    def add_view(self, request, form_url="", extra_context=None):
        """
        Add the SPECTATOR_MAPS setting to context.
        Only currently needed if we're using Mapbox, because we need to set
        the API key in JS in the page.
        """
        extra = extra_context or {}
        extra["SPECTATOR_MAPS"] = app_settings.MAPS
        return super().add_view(request, form_url, extra_context=extra)

    def change_view(self, request, object_id, form_url="", extra_context=None):
        """
        Add the SPECTATOR_MAPS setting to context.
        Only currently needed if we're using Mapbox, because we need to set
        the API key in JS in the page.
        """
        extra = extra_context or {}
        extra["SPECTATOR_MAPS"] = app_settings.MAPS
        return super().change_view(request, object_id, form_url, extra_context=extra)



# Generated by Django 2.0 on 2018-01-25 13:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0010_auto_20180118_0906"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="classicalwork",
            options={"ordering": ("title_sort",), "verbose_name": "classical work"},
        ),
        migrations.AlterModelOptions(
            name="classicalworkrole",
            options={
                "ordering": ("role_order", "role_name"),
                "verbose_name": "classical work role",
            },
        ),
        migrations.AlterModelOptions(
            name="dancepiece",
            options={"ordering": ("title_sort",), "verbose_name": "dance piece"},
        ),
        migrations.AlterModelOptions(
            name="dancepiecerole",
            options={
                "ordering": ("role_order", "role_name"),
                "verbose_name": "dance piece role",
            },
        ),
        migrations.AlterModelOptions(
            name="eventrole",
            options={
                "ordering": ("role_order", "role_name"),
                "verbose_name": "event role",
            },
        ),
        migrations.AlterModelOptions(
            name="movierole",
            options={
                "ordering": ("role_order", "role_name"),
                "verbose_name": "movie role",
            },
        ),
        migrations.AlterModelOptions(
            name="playrole",
            options={
                "ordering": ("role_order", "role_name"),
                "verbose_name": "play role",
            },
        ),
    ]



import factory

from spectator.core.factories import IndividualCreatorFactory

from . import models


class PublicationSeriesFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.PublicationSeries

    title = factory.Sequence(lambda n: f"Publication Series {n}")


class PublicationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Publication

    title = factory.Sequence(lambda n: f"Publication {n}")
    series = factory.SubFactory(PublicationSeriesFactory)
    # Bigger width/height than default detail_thumbnail_2x size:
    thumbnail = factory.django.ImageField(color="blue", width=800, height=800)


class PublicationRoleFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.PublicationRole

    role_name = factory.Sequence(lambda n: f"Role {n}")
    creator = factory.SubFactory(IndividualCreatorFactory)
    publication = factory.SubFactory(PublicationFactory)


class ReadingFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Reading

    publication = factory.SubFactory(PublicationFactory)



from django.urls import path, re_path

from . import views
from .models import Event, Work

app_name = "events"

# Will be like 'movies|plays|concerts' etc:
event_kind_slugs = "|".join(Event.get_valid_kind_slugs())

# Will be like 'classical-works|dance-pieces' etc:
work_kind_slugs = "|".join(Work.get_valid_kind_slugs())

urlpatterns = [
    path("", view=views.EventListView.as_view(), name="home"),
    re_path(
        rf"^types/(?P<kind_slug>{event_kind_slugs})/$",
        view=views.EventListView.as_view(),
        name="event_list",
    ),
    path(
        "venues/",
        view=views.VenueListView.as_view(),
        name="venue_list",
    ),
    re_path(
        r"^venues/(?P<slug>[\w-]+)/$",
        view=views.VenueDetailView.as_view(),
        name="venue_detail",
    ),
    re_path(
        r"^(?P<year>[0-9]{4})/$",
        view=views.EventYearArchiveView.as_view(),
        name="event_year_archive",
    ),
    re_path(
        rf"^(?P<kind_slug>{work_kind_slugs})/$",
        view=views.WorkListView.as_view(),
        name="work_list",
    ),
    re_path(
        rf"^(?P<kind_slug>{work_kind_slugs})/(?P<slug>[\w-]+)/$",
        view=views.WorkDetailView.as_view(),
        name="work_detail",
    ),
    re_path(
        r"^(?P<slug>[\w-]+)/$",
        view=views.EventDetailView.as_view(),
        name="event_detail",
    ),
]



from django.contrib import admin
from imagekit.admin import AdminThumbnail

from .models import Publication, PublicationRole, PublicationSeries, Reading


class ReadingInline(admin.TabularInline):
    model = Reading
    fields = (
        "publication",
        "start_date",
        "end_date",
        "is_finished",
        "start_granularity",
        "end_granularity",
    )
    raw_id_fields = ("publication",)
    extra = 1


class PublicationRoleInline(admin.TabularInline):
    model = PublicationRole
    fields = ("creator", "role_name", "role_order")
    raw_id_fields = ("creator",)
    extra = 1


@admin.register(PublicationSeries)
class PublicationSeriesAdmin(admin.ModelAdmin):
    list_display = ("title",)

    fieldsets = (
        (None, {"fields": ("title", "title_sort", "slug", "url")}),
        (
            "Times",
            {"classes": ("collapse",), "fields": ("time_created", "time_modified")},
        ),
    )

    readonly_fields = ("title_sort", "slug", "time_created", "time_modified")


class ReadingsListFilter(admin.SimpleListFilter):
    """
    Add filters for publications to list 'Unread' and 'In-progress' publications.
    """

    title = "reading"
    parameter_name = "readings"

    def lookups(self, request, model_admin):
        return (("in-progress", ("In progress")), ("unread", ("Unread")))

    def queryset(self, request, queryset):
        if self.value() == "in-progress":
            return queryset.filter(
                reading__start_date__isnull=False, reading__end_date__isnull=True
            )

        if self.value() == "unread":
            return queryset.filter(reading__isnull=True)


@admin.register(Publication)
class PublicationAdmin(admin.ModelAdmin):
    list_display = ("title", "list_thumbnail", "kind", "show_creators", "series")
    list_filter = (ReadingsListFilter, "kind", "series")
    search_fields = ("title",)
    list_select_related = ("series",)

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "title",
                    "title_sort",
                    "detail_thumbnail",
                    "thumbnail",
                    "slug",
                    "kind",
                    "series",
                    "isbn_uk",
                    "isbn_us",
                    "official_url",
                    "notes_url",
                )
            },
        ),
        (
            "Times",
            {"classes": ("collapse",), "fields": ("time_created", "time_modified")},
        ),
    )

    radio_fields = {"kind": admin.HORIZONTAL}
    readonly_fields = (
        "title_sort",
        "detail_thumbnail",
        "slug",
        "time_created",
        "time_modified",
    )

    inlines = [PublicationRoleInline, ReadingInline]

    def show_creators(self, instance):
        names = [str(r.creator) for r in instance.roles.all()]
        if names:
            return ", ".join(names)
        else:
            return "-"

    show_creators.short_description = "Creators"

    detail_thumbnail = AdminThumbnail(
        image_field="thumbnail", template="spectator_core/admin/detail_thumbnail.html"
    )

    list_thumbnail = AdminThumbnail(
        image_field="thumbnail", template="spectator_core/admin/list_thumbnail.html"
    )



# Generated by Django 2.0 on 2018-01-31 15:37

from django.db import migrations, models


def forwards(apps, schema_editor):
    """
    Set the venue_name field of all Events that have a Venue.
    """
    Event = apps.get_model("spectator_events", "Event")

    for event in Event.objects.all():
        if event.venue is not None:
            event.venue_name = event.venue.name
            event.save()


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0023_venue_cinema_treasures_id"),
    ]

    operations = [
        migrations.AddField(
            model_name="event",
            name="venue_name",
            field=models.CharField(
                blank=True,
                help_text="The name of the Venue when this event occurred. If left blank, will be set automatically.",  # noqa: E501
                max_length=255,
            ),
        ),
        migrations.RunPython(forwards),
    ]



# Generated by Django 2.2.3 on 2019-07-31 12:34

from django.db import migrations, models

import spectator.events.models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0040_auto_20180417_1721"),
    ]

    operations = [
        migrations.AddField(
            model_name="event",
            name="ticket",
            field=models.ImageField(
                blank=True,
                default="",
                upload_to=spectator.events.models.event_upload_path,
            ),
        ),
    ]



# Generated by Django 3.0.5 on 2020-04-07 10:50

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_reading", "0006_publication_cover"),
    ]

    operations = [
        migrations.RenameField(
            model_name="publication",
            old_name="cover",
            new_name="thumbnail",
        ),
    ]



from django.db import models
from django.db.models import Min


class InProgressPublicationsManager(models.Manager):
    """
    Returns Publications that are currently being read, ordered with the
    most-recently-started last.
    They might have previously been finished.
    """

    def get_queryset(self):
        from .models import Publication  # noqa: F401

        return (
            super()
            .get_queryset()
            .filter(reading__start_date__isnull=False, reading__end_date__isnull=True)
            .annotate(min_start_date=Min("reading__start_date"))
            .order_by("min_start_date")
        )


class UnreadPublicationsManager(models.Manager):
    """
    Returns Publications that haven't been started (have no Readings).
    """

    def get_queryset(self):
        return super().get_queryset().filter(reading__isnull=True)


class EndDateAscendingReadingsManager(models.Manager):
    """
    Returns Readings in descending end_date order, with Readings that have
    no end_date first.
    Via http://stackoverflow.com/a/15125261/250962
    """

    def get_queryset(self):
        qs = super().get_queryset()
        qs = qs.extra(select={"end_date_null": "end_date is null"})
        return qs.extra(order_by=["end_date_null", "end_date"])


class EndDateDescendingReadingsManager(models.Manager):
    """
    Returns Readings in ascending end_date order, with Readings that have
    no end_date last.
    Via http://stackoverflow.com/a/15125261/250962
    """

    def get_queryset(self):
        qs = super().get_queryset()
        qs = qs.extra(select={"end_date_null": "end_date is null"})
        return qs.extra(order_by=["-end_date_null", "-end_date"])



# Generated by Django 2.0 on 2018-01-25 13:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_reading", "0004_slugs_20180102_1153"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="publicationrole",
            options={
                "ordering": ("role_order", "role_name"),
                "verbose_name": "Publication role",
            },
        ),
        migrations.AlterModelOptions(
            name="publicationseries",
            options={
                "ordering": ("title_sort",),
                "verbose_name": "Publication series",
                "verbose_name_plural": "Publication series",
            },
        ),
    ]



from django.urls import path, re_path

from . import views

app_name = "reading"

urlpatterns = [
    path("", view=views.ReadingHomeView.as_view(), name="home"),
    path(
        "series/",
        view=views.PublicationSeriesListView.as_view(),
        name="publicationseries_list",
    ),
    re_path(
        r"^series/(?P<slug>[\w-]+)/$",
        view=views.PublicationSeriesDetailView.as_view(),
        name="publicationseries_detail",
    ),
    path(
        "publications/",
        view=views.PublicationListView.as_view(),
        name="publication_list",
    ),
    path(
        "publications/periodicals/",
        view=views.PublicationListView.as_view(),
        name="publication_list_periodical",
        kwargs={"kind": "periodical"},
    ),
    re_path(
        r"^publications/(?P<slug>[\w-]+)/$",
        view=views.PublicationDetailView.as_view(),
        name="publication_detail",
    ),
    re_path(
        r"^(?P<year>[0-9]{4})/$",
        view=views.ReadingYearArchiveView.as_view(),
        name="reading_year_archive",
    ),
    re_path(
        r"^(?P<year>[0-9]{4})/(?P<kind>[\w-]+)/$",
        view=views.ReadingYearArchiveView.as_view(),
        name="reading_year_archive",
    ),
]



from urllib.parse import urlparse

from django import template
from django.http import QueryDict
from django.urls import reverse
from django.utils.html import format_html

from spectator.core.apps import spectator_apps
from spectator.core.models import Creator
from spectator.core.utils import chartify

if spectator_apps.is_enabled("events"):
    from spectator.events.models import Venue


register = template.Library()


@register.simple_tag
def get_enabled_apps():
    """
    Returns a list of the optional Spectator apps that are currently installed
    and enabled. e.g. `['events', 'reading',]`.
    """
    return spectator_apps.enabled()


@register.filter
def get_item(dictionary, key):
    """
    For getting an item from a dictionary in a template using a variable.
    Use like:
        {{ mydict|get_item:my_var }}
    """
    return dictionary.get(key)


@register.filter
def get_attr(obj, attr_name):
    """
    For getting a value from a property on an object.
    Use like:
        {{ myobj|get_attr:my_attr_name }}
    """
    return getattr(obj, attr_name, "")


@register.inclusion_tag("spectator_core/includes/card_change_object_link.html")
def change_object_link_card(obj, perms):
    """
    If the user has permission to change `obj`, show a link to its Admin page.
    obj -- An object like Movie, Play, ClassicalWork, Publication, etc.
    perms -- The `perms` object that it's the template.
    """
    # eg: 'movie' or 'classicalwork':
    name = obj.__class__.__name__.lower()
    permission = f"spectator.can_edit_{name}"
    # eg: 'admin:events_classicalwork_change':
    change_url_name = f"admin:{obj._meta.app_label}_{name}_change"

    return {
        "display_link": (permission in perms),
        "change_url": reverse(change_url_name, args=[obj.id]),
    }


@register.filter
def domain_urlize(value):
    """
    Returns an HTML link to the supplied URL, but only using the domain as the
    text. Strips 'www.' from the start of the domain, if present.

    e.g. if `my_url` is 'http://www.example.org/foo/' then:

        {{ my_url|domain_urlize }}

    returns:
        <a href="http://www.example.org/foo/" rel="nofollow">example.org</a>
    """
    parsed_uri = urlparse(value)
    domain = f"{parsed_uri.netloc}"

    if domain.startswith("www."):
        domain = domain[4:]

    return format_html('<a href="{}" rel="nofollow">{}</a>', value, domain)


@register.simple_tag(takes_context=True)
def current_url_name(context):
    """
    Returns the name of the current URL, namespaced, or False.

    Example usage:

        {% current_url_name as url_name %}

        <a href="#"{% if url_name == 'myapp:home' %} class="active"{% endif %}">Home</a>

    """
    url_name = False
    if context.request.resolver_match:
        namespace = context.request.resolver_match.namespace
        name = context.request.resolver_match.url_name
        url_name = f"{namespace}:{name}"
    return url_name


@register.simple_tag(takes_context=True)
def query_string(context, key, value):
    """
    For adding/replacing a key=value pair to the GET string for a URL.

    eg, if we're viewing ?p=3 and we do {% query_string order 'taken' %}
    then this returns "p=3&order=taken"

    And, if we're viewing ?p=3&order=uploaded and we do the same thing, we get
    the same result (ie, the existing "order=uploaded" is replaced).

    Expects the request object in context to do the above; otherwise it will
    just return a query string with the supplied key=value pair.
    """
    try:
        request = context["request"]
        args = request.GET.copy()
    except KeyError:
        args = QueryDict("").copy()
    args[key] = value
    return args.urlencode()


@register.simple_tag
def most_read_creators(num=10):
    """
    Returns a QuerySet of the Creators who have the most Readings associated
    with their Publications.

    Because we're after "most read" we'll only include Creators whose role
    was left empty or is 'Author'.
    """
    return Creator.objects.by_readings()[:num]


@register.inclusion_tag("spectator_core/includes/card_chart.html")
def most_read_creators_card(num=10):
    """
    Displays a card showing the Creators who have the most Readings
    associated with their Publications.

    In spectator_core tags, rather than spectator_reading so it can still be
    used on core pages, even if spectator_reading isn't installed.
    """
    if spectator_apps.is_enabled("reading"):
        object_list = most_read_creators(num=num)

        object_list = chartify(object_list, "num_readings", cutoff=1)

        return {
            "card_title": "Most read authors",
            "score_attr": "num_readings",
            "object_list": object_list,
        }


@register.simple_tag
def most_visited_venues(num=10):
    """
    Returns a QuerySet of the Venues that have the most Events.
    """
    return Venue.objects.by_visits()[:num]


@register.inclusion_tag("spectator_core/includes/card_chart.html")
def most_visited_venues_card(num=10):
    """
    Displays a card showing the Venues that have the most Events.

    In spectator_core tags, rather than spectator_events so it can still be
    used on core pages, even if spectator_events isn't installed.
    """
    if spectator_apps.is_enabled("events"):
        object_list = most_visited_venues(num=num)

        object_list = chartify(object_list, "num_visits", cutoff=1)

        return {
            "card_title": "Most visited venues",
            "score_attr": "num_visits",
            "object_list": object_list,
        }



from django.apps import AppConfig


class SpectatorReadingAppConfig(AppConfig):
    label = "spectator_reading"
    name = "spectator.reading"
    verbose_name = "Spectator Reading"

    # Maintain pre Django 3.2 default behaviour:
    default_auto_field = "django.db.models.AutoField"



# Generated by Django 2.0 on 2018-01-25 17:57

from django.db import migrations, models


class Migration(migrations.Migration):
    """
    Now we've deleted the old classicalworks and dancepieces fields,
    rename the new ones to be the same as the old.
    """

    dependencies = [
        ("spectator_events", "0014_remove_old_classical_dance"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="event",
            name="classicalworks2",
        ),
        migrations.RemoveField(
            model_name="event",
            name="dancepieces2",
        ),
        migrations.AddField(
            model_name="event",
            name="classicalworks",
            field=models.ManyToManyField(
                blank=True,
                help_text="Only used if event is of 'Classical Concert' kind.",
                through="spectator_events.ClassicalWorkSelection",
                to="spectator_events.ClassicalWork",
            ),
        ),
        migrations.AddField(
            model_name="event",
            name="dancepieces",
            field=models.ManyToManyField(
                blank=True,
                help_text="Only used if event is of 'Dance' kind.",
                through="spectator_events.DancePieceSelection",
                to="spectator_events.DancePiece",
            ),
        ),
    ]



# Generated by Django 2.0 on 2018-04-17 17:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0039_populate_exhibitions"),
    ]

    operations = [
        migrations.AlterField(
            model_name="work",
            name="kind",
            field=models.CharField(
                choices=[
                    ("classicalwork", "Classical work"),
                    ("dancepiece", "Dance piece"),
                    ("exhibition", "Exhibition"),
                    ("movie", "Movie"),
                    ("play", "Play"),
                ],
                max_length=20,
            ),
        ),
    ]



# Generated by Django 2.0 on 2018-04-17 12:18

from django.db import migrations


def forwards(apps, schema_editor):
    """
    Migrate all 'exhibition' Events to the new 'museum' Event kind.
    """
    Event = apps.get_model("spectator_events", "Event")

    for ev in Event.objects.filter(kind="exhibition"):
        ev.kind = "museum"
        ev.save()


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0036_auto_20180417_1218"),
    ]

    operations = [
        migrations.RunPython(forwards),
    ]



# Generated by Django 3.0.5 on 2020-04-07 10:51

from django.db import migrations, models

import spectator.core.models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0043_rename_ticket_to_thumbnail"),
    ]

    operations = [
        migrations.AlterField(
            model_name="event",
            name="thumbnail",
            field=models.ImageField(
                blank=True,
                default="",
                upload_to=spectator.core.models.thumbnail_upload_path,
            ),
        ),
    ]



# Generated by Django 2.0 on 2018-02-08 11:26

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models

import spectator.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_core", "0004_auto_20180102_0959"),
        ("spectator_events", "0025_auto_20180131_1755"),
    ]

    operations = [
        migrations.CreateModel(
            name="Work",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The time this item was created in the database.",
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The time this item was last saved to the database.",
                    ),
                ),
                ("slug", models.SlugField(blank=True, max_length=10)),
                (
                    "kind",
                    models.CharField(
                        choices=[
                            ("classicalwork", "Classical work"),
                            ("dancepiece", "Dance piece"),
                            ("movie", "Movie"),
                            ("play", "Play"),
                        ],
                        max_length=20,
                    ),
                ),
                ("title", models.CharField(max_length=255)),
                (
                    "title_sort",
                    spectator.core.fields.NaturalSortField(
                        "title",
                        db_index=True,
                        default="",
                        editable=False,
                        help_text="e.g. 'big piece, a' or 'biggest piece, the'.",
                        max_length=255,
                    ),
                ),
                (
                    "imdb_id",
                    models.CharField(
                        blank=True,
                        help_text="Starts with 'tt', e.g. 'tt0100842'.",
                        max_length=12,
                        validators=[
                            django.core.validators.RegexValidator(
                                code="invalid_imdb_id",
                                message='IMDb ID should be like "tt1234567"',
                                regex="^tt\\d{7,10}$",
                            )
                        ],
                        verbose_name="IMDb ID",
                    ),
                ),
                (
                    "year",
                    models.PositiveSmallIntegerField(
                        blank=True,
                        default=None,
                        help_text="Year of release, composition, publication, etc.",
                        null=True,
                    ),
                ),
            ],
            options={"verbose_name": "work", "ordering": ("title_sort",)},
        ),
        migrations.CreateModel(
            name="WorkRole",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The time this item was created in the database.",
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The time this item was last saved to the database.",
                    ),
                ),
                (
                    "role_name",
                    models.CharField(
                        blank=True,
                        help_text="e.g. 'Headliner', 'Support', 'Editor', 'Illustrator', 'Director', etc. Optional.",  # noqa: E501
                        max_length=50,
                    ),
                ),
                (
                    "role_order",
                    models.PositiveSmallIntegerField(
                        default=1,
                        help_text="The order in which the Creators will be listed.",
                    ),
                ),
                (
                    "creator",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="work_roles",
                        to="spectator_core.Creator",
                    ),
                ),
                (
                    "work",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="roles",
                        to="spectator_events.Work",
                    ),
                ),
            ],
            options={
                "verbose_name": "work role",
                "ordering": ("role_order", "role_name"),
            },
        ),
        migrations.CreateModel(
            name="WorkSelection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "order",
                    models.PositiveSmallIntegerField(
                        default=1, help_text="Position on the Event programme."
                    ),
                ),
                (
                    "event",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="work_selections",
                        to="spectator_events.Event",
                    ),
                ),
                (
                    "work",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="events",
                        to="spectator_events.Work",
                    ),
                ),
            ],
            options={"verbose_name": "work selection", "ordering": ("order",)},
        ),
        migrations.AddField(
            model_name="work",
            name="creators",
            field=models.ManyToManyField(
                related_name="works",
                through="spectator_events.WorkRole",
                to="spectator_core.Creator",
            ),
        ),
    ]



# Generated by Django 2.0 on 2018-02-08 11:47

from django.db import migrations


def forwards(apps, schema_editor):
    """
    Change all Play objects into Work objects, and their associated
    data into WorkRole and WorkSelection models, then delete the Play.
    """
    Play = apps.get_model("spectator_events", "Play")
    Work = apps.get_model("spectator_events", "Work")
    WorkRole = apps.get_model("spectator_events", "WorkRole")
    WorkSelection = apps.get_model("spectator_events", "WorkSelection")

    for p in Play.objects.all():

        work = Work.objects.create(kind="play", title=p.title, title_sort=p.title_sort)

        for role in p.roles.all():
            WorkRole.objects.create(
                creator=role.creator,
                work=work,
                role_name=role.role_name,
                role_order=role.role_order,
            )

        for selection in p.events.all():
            WorkSelection.objects.create(
                event=selection.event, work=work, order=selection.order
            )

        p.delete()


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0028_dancepieces_to_works"),
    ]

    operations = [
        migrations.RunPython(forwards),
    ]



from django.db.models import Min
from django.http import Http404
from django.utils.encoding import force_str
from django.utils.translation import gettext as _
from django.views.generic import DetailView, YearArchiveView
from django.views.generic.detail import SingleObjectMixin

from spectator.core import app_settings
from spectator.core.views import PaginatedListView

from .models import Event, Venue, Work


class EventListView(PaginatedListView):
    """
    Includes context of counts of all different Event types,
    plus the kind of event this page is for,
    plus adding `event_list` (synonym for `object_list`).

    Expects a `kind_slug` like 'movies', 'gigs', 'concerts', etc.
    """

    model = Event
    ordering = [
        "-date",
    ]

    def get(self, request, *args, **kwargs):
        slug = self.kwargs.get("kind_slug", None)
        if slug is not None and slug not in Event.get_valid_kind_slugs():
            msg = f"Invalid kind_slug: '{slug}'"
            raise Http404(msg)

        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context.update(self.get_event_counts())

        # e.g. 'concert'
        kind = self.get_event_kind()
        context["event_kind"] = kind

        if kind:
            # e.g. 'Concert':
            context["event_kind_name"] = Event.get_kind_name(kind)
            # e.g. 'Concerts':
            context["event_kind_name_plural"] = Event.get_kind_name_plural(kind)

        context["event_list"] = context["object_list"]

        return context

    def get_event_counts(self):
        """
        Returns a dict like:
            {'counts': {
                'all': 30,
                'movie': 12,
                'gig': 10,
            }}
        """
        counts = {
            "all": Event.objects.count(),
        }

        for val in Event.Kind.values:
            # e.g. 'movie_count':
            counts[val] = Event.objects.filter(kind=val).count()

        return {
            "counts": counts,
        }

    def get_event_kind(self):
        """
        Unless we're on the front page we'll have a kind_slug like 'movies'.
        We need to translate that into an event `kind` like 'movie'.
        """
        slug = self.kwargs.get("kind_slug", None)
        if slug is None:
            return None  # Front page; showing all Event kinds.
        else:
            slugs_to_kinds = {v: k for k, v in Event.Kind.slugs().items()}
            return slugs_to_kinds.get(slug)

    def get_queryset(self):
        "Restrict to a single kind of event, if any, and include Venue data."
        qs = super().get_queryset()

        kind = self.get_event_kind()
        if kind is not None:
            qs = qs.filter(kind=kind)

        qs = qs.select_related("venue")

        return qs


class EventDetailView(DetailView):
    model = Event


class EventYearArchiveView(YearArchiveView):
    allow_empty = True
    date_field = "date"
    make_object_list = True
    model = Event
    ordering = "date"

    def get_queryset(self):
        "Reduce the number of queries and speed things up."
        qs = super().get_queryset()
        qs = qs.select_related("venue")
        return qs

    def get_dated_items(self):
        items, qs, info = super().get_dated_items()

        if "year" in info and info["year"]:
            # Get the earliest date we have an Event for:
            date_min = Event.objects.aggregate(Min("date"))["date__min"]
            # Make it a 'yyyy-01-01' date:
            min_year_date = date_min.replace(month=1, day=1)
            if info["year"] < min_year_date:
                # The year we're viewing is before our minimum date, so 404.
                raise Http404(
                    _("No %(verbose_name_plural)s available")
                    % {
                        "verbose_name_plural": force_str(
                            qs.model._meta.verbose_name_plural
                        )
                    }
                )
            elif info["year"] == min_year_date:
                # This is the earliest year we have events for, so
                # there is no previous year.
                info["previous_year"] = None

        return items, qs, info


# WORKS


class WorkMixin:
    kind_slug = None

    def get(self, request, *args, **kwargs):
        slug = self.kwargs.get("kind_slug", None)
        if slug is not None and slug not in Work.get_valid_kind_slugs():
            msg = f"Invalid kind_slug: '{slug}'"
            raise Http404(msg)
        else:
            self.kind_slug = slug

        return super().get(request, *args, **kwargs)

    def get_work_kind(self):
        """
        We'll have a kind_slug like 'movies'.
        We need to translate that into a work `kind` like 'movie'.
        """
        slugs_to_kinds = {v: k for k, v in Work.Kind.slugs().items()}
        return slugs_to_kinds.get(self.kind_slug, None)


class WorkListView(WorkMixin, PaginatedListView):
    model = Work

    def get_queryset(self):
        kind = self.get_work_kind()
        qs = super().get_queryset()
        qs = qs.filter(kind=kind)
        qs = qs.prefetch_related("roles__creator")
        return qs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # 'movie', 'Movie', and 'Movies' respectively:
        kind = self.get_work_kind()
        kind_name = Work.get_kind_name(kind)
        kind_name_plural = Work.get_kind_name_plural(kind)

        context["page_title"] = kind_name_plural
        context["breadcrumb_list_title"] = kind_name_plural

        context["work_kind"] = kind
        context["work_kind_name"] = kind_name
        context["work_kind_name_plural"] = kind_name_plural

        context["breadcrumb_list_url"] = self.model().get_list_url(
            kind_slug=self.kind_slug
        )

        return context


class WorkDetailView(WorkMixin, DetailView):
    model = Work

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        kind = self.get_work_kind()

        context["breadcrumb_list_title"] = Work.get_kind_name_plural(kind)

        context["breadcrumb_list_url"] = self.model().get_list_url(
            kind_slug=self.kind_slug
        )

        return context


# VENUES


class VenueListView(PaginatedListView):
    model = Venue
    ordering = ["name_sort"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context["country_list"] = self.get_countries()

        return context

    def get_countries(self):
        """
        Returns a list of dicts, one per country that has at least one Venue
        in it.

        Each dict has 'code' and 'name' elements.
        The list is sorted by the country 'name's.
        """
        qs = (
            Venue.objects.values("country")
            .exclude(country="")
            .distinct()
            .order_by("country")
        )

        countries = []

        for c in qs:
            countries.append(
                {"code": c["country"], "name": Venue.get_country_name(c["country"])}
            )

        return sorted(countries, key=lambda k: k["name"])


class VenueDetailView(SingleObjectMixin, PaginatedListView):
    template_name = "spectator_events/venue_detail.html"

    def get(self, request, *args, **kwargs):
        self.object = self.get_object(queryset=Venue.objects.all())
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context["venue"] = self.object

        context["event_list"] = context["object_list"]

        if self.object.latitude is None or self.object.longitude is None:
            context["SPECTATOR_MAPS"] = {"enable": False}
        else:
            context["SPECTATOR_MAPS"] = app_settings.MAPS

        return context

    def get_queryset(self):
        return self.object.event_set.order_by("-date")



# Generated by Django 1.11 on 2017-04-17 17:14

import django.db.models.deletion
from django.db import migrations, models

import spectator.core.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("spectator_core", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Publication",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The time this item was created in the database.",
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The time this item was last saved to the database.",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="e.g. 'Aurora' or 'Vol. 39 No. 4, 16 February 2017'.",
                        max_length=255,
                    ),
                ),
                (
                    "title_sort",
                    spectator.core.fields.NaturalSortField(
                        "title",
                        db_index=True,
                        default="",
                        editable=False,
                        help_text="e.g. 'clockwork orange, a' or 'world cities, the'.",
                        max_length=255,
                    ),
                ),
                (
                    "kind",
                    models.CharField(
                        choices=[("book", "Book"), ("periodical", "Periodical")],
                        default="book",
                        max_length=20,
                    ),
                ),
                (
                    "official_url",
                    models.URLField(
                        blank=True,
                        help_text="Official URL for this book/issue.",
                        max_length=255,
                        verbose_name="Official URL",
                    ),
                ),
                (
                    "isbn_uk",
                    models.CharField(
                        blank=True,
                        help_text="e.g. '0356500489'.",
                        max_length=20,
                        verbose_name="UK ISBN",
                    ),
                ),
                (
                    "isbn_us",
                    models.CharField(
                        blank=True,
                        help_text="e.g. '0316098094'.",
                        max_length=20,
                        verbose_name="US ISBN",
                    ),
                ),
                (
                    "notes_url",
                    models.URLField(
                        blank=True,
                        help_text="URL of your notes/review.",
                        max_length=255,
                        verbose_name="Notes URL",
                    ),
                ),
            ],
            options={"ordering": ("title_sort",)},
        ),
        migrations.CreateModel(
            name="PublicationRole",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The time this item was created in the database.",
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The time this item was last saved to the database.",
                    ),
                ),
                (
                    "role_name",
                    models.CharField(
                        blank=True,
                        help_text="e.g. 'Headliner', 'Support', 'Editor', 'Illustrator', 'Director', etc. Optional.",  # noqa: E501
                        max_length=50,
                    ),
                ),
                (
                    "role_order",
                    models.PositiveSmallIntegerField(
                        default=1,
                        help_text="The order in which the Creators will be listed.",
                    ),
                ),
                (
                    "creator",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="publication_roles",
                        to="spectator_core.Creator",
                    ),
                ),
                (
                    "publication",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="roles",
                        to="spectator_reading.Publication",
                    ),
                ),
            ],
            options={"ordering": ("role_order", "role_name"), "abstract": False},
        ),
        migrations.CreateModel(
            name="PublicationSeries",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The time this item was created in the database.",
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The time this item was last saved to the database.",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="e.g. 'The London Review of Books'.", max_length=255
                    ),
                ),
                (
                    "title_sort",
                    spectator.core.fields.NaturalSortField(
                        "title",
                        db_index=True,
                        default="",
                        editable=False,
                        help_text="e.g. 'london review of books, the'.",
                        max_length=255,
                    ),
                ),
                (
                    "url",
                    models.URLField(
                        blank=True,
                        help_text="e.g. 'https://www.lrb.co.uk/'.",
                        max_length=255,
                        verbose_name="URL",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Publication series",
                "ordering": ("title_sort",),
            },
        ),
        migrations.CreateModel(
            name="Reading",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The time this item was created in the database.",
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The time this item was last saved to the database.",
                    ),
                ),
                ("start_date", models.DateField(blank=True, null=True)),
                (
                    "start_granularity",
                    models.PositiveSmallIntegerField(
                        choices=[(3, "Y-m-d"), (4, "Y-m"), (6, "Y")], default=3
                    ),
                ),
                ("end_date", models.DateField(blank=True, null=True)),
                (
                    "end_granularity",
                    models.PositiveSmallIntegerField(
                        choices=[(3, "Y-m-d"), (4, "Y-m"), (6, "Y")], default=3
                    ),
                ),
                (
                    "is_finished",
                    models.BooleanField(
                        default=False, help_text="Did you finish the publication?"
                    ),
                ),
                (
                    "publication",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="spectator_reading.Publication",
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.AddField(
            model_name="publication",
            name="creators",
            field=models.ManyToManyField(
                related_name="publications",
                through="spectator_reading.PublicationRole",
                to="spectator_core.Creator",
            ),
        ),
        migrations.AddField(
            model_name="publication",
            name="series",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="spectator_reading.PublicationSeries",
            ),
        ),
    ]



__title__ = "Django Spectator"
__version__ = "15.2.0"
__author__ = "Phil Gyford"
__author_email__ = "phil@gyford.com"
__license__ = "MIT"

VERSION = __version__  # synonym
TITLE = __title__  # synonym



import logging

from django.conf import settings

LOGGER = logging.getLogger(__name__)

# Creating all the defaults for settings.
# In our code, if we want to use a SPECTATOR_* setting we should import
# from here, not django.conf.settings.

if getattr(settings, "SPECTATOR_GOOGLE_MAPS_API_KEY", False):
    LOGGER.warning(
        "The SPECTATOR_GOOGLE_MAPS_API_KEY setting is no longer recognised. "
        "Please update to use the SPECTATOR_MAPS dictionary setting."
    )


MAPS = getattr(settings, "SPECTATOR_MAPS", {"enable": False})


# The characters to use for things that require an automatically-generated
# URL slug:
SLUG_ALPHABET = getattr(
    settings, "SPECTATOR_SLUG_ALPHABET", "abcdefghijkmnopqrstuvwxyz23456789"
)

# The salt value to use when generating URL slugs:
SLUG_SALT = getattr(settings, "SPECTATOR_SLUG_SALT", "Django Spectator")

# For Events and card titles.
DATE_FORMAT = getattr(settings, "SPECTATOR_DATE_FORMAT", "%-d %b %Y")

# For Reading and Events:
THUMBNAIL_DETAIL_SIZE = getattr(settings, "SPECTATOR_THUMBNAIL_DETAIL_SIZE", (320, 320))
THUMBNAIL_LIST_SIZE = getattr(settings, "SPECTATOR_THUMBNAIL_LIST_SIZE", (80, 160))

# Top-level directories, within MEDIA_ROOT, for the Event and
# Publication thumbnails to go in:
EVENTS_DIR_BASE = getattr(settings, "SPECTATOR_EVENTS_DIR_BASE", "events")
READING_DIR_BASE = getattr(settings, "SPECTATOR_READING_DIR_BASE", "reading")



import factory

from . import models


class IndividualCreatorFactory(factory.django.DjangoModelFactory):
    "A creator that is an individual person."

    class Meta:
        model = models.Creator

    name = factory.Sequence(lambda n: f"Individual {n}")
    kind = "individual"


class GroupCreatorFactory(factory.django.DjangoModelFactory):
    "A creator that is a group/organisation/etc."

    class Meta:
        model = models.Creator

    name = factory.Sequence(lambda n: f"Group {n}")
    kind = "group"



# Generated by Django 1.11 on 2017-11-01 16:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="event",
            name="slug",
            field=models.SlugField(blank=True, default="a", max_length=10),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.0 on 2018-01-25 17:39

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0011_auto_20180125_1348"),
    ]

    operations = [
        migrations.CreateModel(
            name="ClassicalWorkSelection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "order",
                    models.PositiveSmallIntegerField(
                        default=1, help_text="Position on the Event programme."
                    ),
                ),
                (
                    "classical_work",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="events",
                        to="spectator_events.ClassicalWork",
                    ),
                ),
                (
                    "event",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="classical_work_selections",
                        to="spectator_events.Event",
                    ),
                ),
            ],
            options={
                "verbose_name": "classical work selection",
                "ordering": ("order",),
            },
        ),
        migrations.CreateModel(
            name="DancePieceSelection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "order",
                    models.PositiveSmallIntegerField(
                        default=1, help_text="Position on the Event programme."
                    ),
                ),
                (
                    "dance_piece",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="events",
                        to="spectator_events.DancePiece",
                    ),
                ),
                (
                    "event",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dance_piece_selections",
                        to="spectator_events.Event",
                    ),
                ),
            ],
            options={"verbose_name": "dance piece selection", "ordering": ("order",)},
        ),
        migrations.AddField(
            model_name="event",
            name="classicalworks2",
            field=models.ManyToManyField(
                blank=True,
                help_text="Only used if event is of 'Classical Concert' kind.",
                related_name="event_set_2",
                through="spectator_events.ClassicalWorkSelection",
                to="spectator_events.ClassicalWork",
            ),
        ),
        migrations.AddField(
            model_name="event",
            name="dancepieces2",
            field=models.ManyToManyField(
                blank=True,
                help_text="Only used if event is of 'Dance' kind.",
                related_name="event_set_2",
                through="spectator_events.DancePieceSelection",
                to="spectator_events.DancePiece",
            ),
        ),
    ]



from django.urls import path

from spectator.core import views

# Only the home page.
# This should be under the namespace 'spectator:core'.

app_name = "core"

urlpatterns = [
    path("", view=views.HomeView.as_view(), name="home"),
]



from django import template
from django.db.models import Count
from django.db.models.functions import TruncYear
from django.utils.html import format_html

from spectator.core import app_settings
from spectator.core.models import Creator
from spectator.core.utils import chartify
from spectator.events.models import Event, Work

register = template.Library()


@register.simple_tag
def annual_event_counts(kind="all"):
    """
    Returns a QuerySet of dicts, each one with these keys:

        * year - a date object representing the year
        * total - the number of events of `kind` that year

    kind - The Event `kind`, or 'all' for all kinds (default).
    """
    qs = Event.objects

    if kind != "all":
        qs = qs.filter(kind=kind)

    qs = (
        qs.annotate(year=TruncYear("date"))
        .values("year")
        .annotate(total=Count("id"))
        .order_by("year")
    )

    return qs


@register.inclusion_tag("spectator_events/includes/card_annual_event_counts.html")
def annual_event_counts_card(kind="all", current_year=None):
    """
    Displays years and the number of events per year.

    kind is an Event kind (like 'cinema', 'gig', etc.) or 'all' (default).
    current_year is an optional date object representing the year we're already
        showing information about.
    """
    if kind == "all":
        card_title = "Events per year"
    else:
        card_title = f"{Event.get_kind_name_plural(kind)} per year"

    return {
        "card_title": card_title,
        "kind": kind,
        "years": annual_event_counts(kind=kind),
        "current_year": current_year,
    }


@register.simple_tag
def display_date(d):
    """
    Render a date/datetime (d) as a date, using the SPECTATOR_DATE_FORMAT
    setting. Wrap the output in a <time> tag.

    Time tags: http://www.brucelawson.co.uk/2012/best-of-time/
    """
    stamp = d.strftime("%Y-%m-%d")
    visible_date = d.strftime(app_settings.DATE_FORMAT)

    return format_html('<time datetime="{}">{}</time>', stamp, visible_date)


@register.inclusion_tag("spectator_events/includes/event_list_tabs.html")
def event_list_tabs(counts, current_kind, page_number=1):
    """
    Displays the tabs to different event_list pages.

    `counts` is a dict of number of events for each kind, like:
        {'all': 30, 'gig': 12, 'movie': 18,}

    `current_kind` is the event kind that's active, if any. e.g. 'gig',
        'movie', etc.

    `page_number` is the current page of this kind of events we're on.
    """
    return {
        "counts": counts,
        "current_kind": current_kind,
        "page_number": page_number,
        # A list of all the kinds we might show tabs for, like
        # ['gig', 'movie', 'play', ...]
        "event_kinds": Event.get_kinds(),
        # A dict of data about each kind, keyed by kind ('gig') including
        # data about 'name', 'name_plural' and 'slug':
        "event_kinds_data": Event.get_kinds_data(),
    }


@register.simple_tag
def recent_events(num=10):
    """
    Returns a QuerySet of Events that happened recently.
    `num` is the number returned.
    """
    return Event.objects.select_related("venue").order_by("-date")[:num]


@register.inclusion_tag("spectator_events/includes/card_events.html")
def recent_events_card(num=10):
    """
    Displays Events that happened recently.
    `num` is the number returned.
    """
    return {
        "card_title": "Recent events",
        "event_list": recent_events(num=num),
    }


@register.simple_tag
def day_events(date):
    """
    Returns a QuerySet of Events that happened on the supplied date.
    `date` is a date object.
    """
    return Event.objects.filter(date=date).select_related("venue")


@register.inclusion_tag("spectator_events/includes/card_events.html")
def day_events_card(date):
    """
    Displays Events that happened on the supplied date.
    `date` is a date object.
    """
    d = date.strftime(app_settings.DATE_FORMAT)
    card_title = f"Events on {d}"
    return {
        "card_title": card_title,
        "event_list": day_events(date=date),
    }


@register.simple_tag
def events_years():
    """
    Returns a QuerySet of date objects, one for each year in which there are
    Events.
    """
    return Event.objects.dates("date", "year")


@register.inclusion_tag("spectator_events/includes/card_years.html")
def events_years_card(current_year=None):
    """
    Displays a card showing all years in which we have Events, with a link to
    the archive pages.

    current_year is a date object representing the year we're displaying.
    """
    return {
        "current_year": current_year,
        "years": events_years(),
    }


@register.simple_tag
def most_seen_creators(event_kind=None, num=10):
    """
    Returns a QuerySet of the Creators that are associated with the most Events.
    """
    return Creator.objects.by_events(kind=event_kind)[:num]


@register.inclusion_tag("spectator_core/includes/card_chart.html")
def most_seen_creators_card(event_kind=None, num=10):
    """
    Displays a card showing the Creators that are associated with the most Events.
    """
    object_list = most_seen_creators(event_kind=event_kind, num=num)

    object_list = chartify(object_list, "num_events", cutoff=1)

    return {
        "card_title": "Most seen people/groups",
        "score_attr": "num_events",
        "object_list": object_list,
    }


@register.simple_tag
def most_seen_creators_by_works(work_kind=None, role_name=None, num=10):
    """
    Returns a QuerySet of the Creators that are associated with the most Works.
    """
    return Creator.objects.by_works(kind=work_kind, role_name=role_name)[:num]


@register.inclusion_tag("spectator_core/includes/card_chart.html")
def most_seen_creators_by_works_card(work_kind=None, role_name=None, num=10):
    """
    Displays a card showing the Creators that are associated with the most Works.

    e.g.:
    {% most_seen_creators_by_works_card work_kind='movie' role_name='Director' num=5 %}
    """
    object_list = most_seen_creators_by_works(
        work_kind=work_kind, role_name=role_name, num=num
    )

    object_list = chartify(object_list, "num_works", cutoff=1)

    # Attempt to create a sensible card title...

    # Yes, this pluralization is going to break at some point:
    creators_name = f"{role_name.capitalize()}s" if role_name else "People/groups"

    works_name = Work.get_kind_name_plural(work_kind).lower() if work_kind else "works"

    card_title = f"{creators_name} with most {works_name}"

    return {
        "card_title": card_title,
        "score_attr": "num_works",
        "object_list": object_list,
    }


@register.simple_tag
def most_seen_works(kind=None, num=10):
    """
    Returns a QuerySet of the Works that are associated with the most Events.
    """
    return Work.objects.by_views(kind=kind)[:num]


@register.inclusion_tag("spectator_core/includes/card_chart.html")
def most_seen_works_card(kind=None, num=10):
    """
    Displays a card showing the Works that are associated with the most Events.
    """
    object_list = most_seen_works(kind=kind, num=num)

    object_list = chartify(object_list, "num_views", cutoff=1)

    if kind:
        card_title = f"Most seen {Work.get_kind_name_plural(kind).lower()}"
    else:
        card_title = "Most seen works"

    return {
        "card_title": card_title,
        "score_attr": "num_views",
        "object_list": object_list,
        "name_attr": "title",
        "use_cite": True,
    }



# Generated by Django 2.0 on 2018-01-29 17:52

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0021_auto_20180129_1735"),
    ]

    operations = [
        migrations.AlterField(
            model_name="event",
            name="venue",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="spectator_events.Venue",
            ),
        ),
    ]



# Generated by Django 2.2.3 on 2019-07-22 16:47

from django.db import migrations, models

import spectator.reading.models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_reading", "0005_auto_20180125_1348"),
    ]

    operations = [
        migrations.AddField(
            model_name="publication",
            name="cover",
            field=models.ImageField(
                blank=True,
                default="",
                upload_to=spectator.reading.models.publication_upload_path,
            ),
        ),
    ]



# Generated by Django 3.0.5 on 2020-04-07 10:39

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0041_event_ticket"),
    ]

    operations = [
        migrations.AlterField(
            model_name="venue",
            name="cinema_treasures_id",
            field=models.PositiveIntegerField(
                blank=True,
                help_text='Optional. ID of a cinema at\n<a href="http://cinematreasures.org/">Cinema Treasures</a>.',  # noqa: E501
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="work",
            name="imdb_id",
            field=models.CharField(
                blank=True,
                help_text="Starts with 'tt', e.g. 'tt0100842'.\nFrom <a href=\"https://www.imdb.com\">IMDb</a>.",  # noqa: E501
                max_length=12,
                validators=[
                    django.core.validators.RegexValidator(
                        code="invalid_imdb_id",
                        message='IMDb ID should be like "tt1234567"',
                        regex="^tt\\d{7,10}$",
                    )
                ],
                verbose_name="IMDb ID",
            ),
        ),
    ]



from django.urls import path, re_path

from spectator.core import views

app_name = "creators"

urlpatterns = [
    # Individuals:
    path("", view=views.CreatorListView.as_view(), name="creator_list"),
    #  Groups:
    path(
        "groups/",
        view=views.CreatorListView.as_view(),
        name="creator_list_group",
        kwargs={"kind": "group"},
    ),
    re_path(
        r"^(?P<slug>[\w-]+)/$",
        view=views.CreatorDetailView.as_view(),
        name="creator_detail",
    ),
]



# Generated by Django 2.0 on 2018-04-17 12:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0035_change_event_kinds"),
    ]

    operations = [
        migrations.AlterField(
            model_name="event",
            name="kind",
            field=models.CharField(
                choices=[
                    ("cinema", "Cinema"),
                    ("concert", "Concert"),
                    ("comedy", "Comedy"),
                    ("dance", "Dance"),
                    ("exhibition", "Exhibition"),
                    ("museum", "Gallery/Museum"),
                    ("gig", "Gig"),
                    ("theatre", "Theatre"),
                    ("misc", "Other"),
                ],
                help_text="Used to categorise event. But any kind of Work can be added to any kind of Event.",  # noqa: E501
                max_length=20,
            ),
        ),
    ]



from django.apps import AppConfig


class SpectatorEventsAppConfig(AppConfig):
    label = "spectator_events"
    name = "spectator.events"
    verbose_name = "Spectator Events"

    # Maintain pre Django 3.2 default behaviour:
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        import spectator.events.signals  # noqa: F401



from django import template
from django.db.models import Q

from spectator.core import app_settings
from spectator.reading import utils
from spectator.reading.models import Publication, Reading

register = template.Library()


@register.simple_tag
def annual_reading_counts(kind="all"):
    """
    Returns a list of dicts, one per year of reading. In year order.
    Each dict is like this (if kind is 'all'):

        {'year':        datetime.date(2003, 1, 1),
         'book':        12,    # only included if kind is 'all' or 'book'
         'periodical':  18,    # only included if kind is 'all' or 'periodical'
         'total':       30,    # only included if kind is 'all'
         }

    We use the end_date of a Reading to count when that thing was read.

    kind is one of 'book', 'periodical' or 'all', for both.
    """
    return utils.annual_reading_counts(kind=kind)


@register.inclusion_tag("spectator_reading/includes/card_annual_reading_counts.html")
def annual_reading_counts_card(kind="all", current_year=None):
    """
    Displays years and the number of books/periodicals read per year.

    kind is one of 'book', 'periodical', 'all' (default).
    current_year is an optional date object representing the year we're already
        showing information about.
    """
    if kind == "book":
        card_title = "Books per year"
    elif kind == "periodical":
        card_title = "Periodicals per year"
    else:
        card_title = "Reading per year"

    return {
        "card_title": card_title,
        "kind": kind,
        "years": utils.annual_reading_counts(kind),
        "current_year": current_year,
    }


@register.simple_tag
def in_progress_publications():
    """
    Returns a QuerySet of any Publications that are currently being read.
    """
    return Publication.in_progress_objects.select_related("series").prefetch_related(
        "roles__creator"
    )


@register.inclusion_tag("spectator_reading/includes/card_publications.html")
def in_progress_publications_card():
    """
    Displays Publications that are currently being read.
    """
    return {
        "card_title": "Currently reading",
        "publication_list": in_progress_publications(),
    }


@register.simple_tag
def day_publications(date):
    """
    Returns a QuerySet of Publications that were being read on `date`.
    `date` is a date tobject.
    """
    readings = Reading.objects.filter(start_date__lte=date).filter(
        Q(end_date__gte=date) | Q(end_date__isnull=True)
    )
    if readings:
        return (
            Publication.objects.filter(reading__in=readings)
            .select_related("series")
            .prefetch_related("roles__creator")
            .distinct()
        )
    else:
        return Publication.objects.none()


@register.inclusion_tag("spectator_reading/includes/card_publications.html")
def day_publications_card(date):
    """
    Displays Publications that were being read on `date`.
    `date` is a date tobject.
    """
    d = date.strftime(app_settings.DATE_FORMAT)
    card_title = f"Reading on {d}"
    return {
        "card_title": card_title,
        "publication_list": day_publications(date=date),
    }


@register.simple_tag
def reading_years():
    """
    Returns a QuerySet of date objects, one for each year in which there are
    Readings.
    """
    return Reading.objects.dates("end_date", "year")


@register.inclusion_tag("spectator_reading/includes/card_years.html")
def reading_years_card(current_year=None):
    """
    Displays the years in which there are Readings.
    Each one is linked to the year archive page.
    current_year is a date object; this year won't be linked.
    """
    return {
        "current_year": current_year,
        "years": reading_years(),
    }


@register.simple_tag
def reading_dates(reading):
    """
    Given a Reading, with start and end dates and granularities[1] it returns
    an HTML string representing that period. eg:

        * '1–6 Feb 2017'
        * '1 Feb to 3 Mar 2017'
        * 'Feb 2017 to Mar 2018'
        * '2017–2018'
    etc.

    [1] https://www.flickr.com/services/api/misc.dates.html
    """

    # 3 September 2017
    full_format = '<time datetime="%Y-%m-%d">{}</time>'.format("%-d %B %Y")
    # September 2017
    month_year_format = '<time datetime="%Y-%m">{}</time>'.format("%B %Y")
    # 2017
    year_format = '<time datetime="%Y">{}</time>'.format("%Y")
    # 3
    day_format = '<time datetime="%Y-%m-%d">{}</time>'.format("%-d")
    # 3 September
    day_month_format = '<time datetime="%Y-%m-%d">{}</time>'.format("%-d %B")
    # September
    month_format = '<time datetime="%Y-%m">{}</time>'.format("%B")

    period_format_short = "{}–{}"

    period_format_long = "{} to {}"

    # For brevity:
    start_date = reading.start_date
    end_date = reading.end_date
    start_gran = reading.start_granularity
    end_gran = reading.end_granularity

    # Are start and end in the same day, year or month?
    same_day = False
    same_month = False
    same_year = False

    if start_date and end_date and start_date.strftime("%Y") == end_date.strftime("%Y"):
        same_year = True
        if start_date.strftime("%m%Y") == end_date.strftime("%m%Y"):
            same_month = True
            if start_date.strftime("%d%m%Y") == end_date.strftime("%d%m%Y"):
                same_day = True

    start_str = ""
    end_str = ""
    output = ""

    # Make some basic start and end strings, which we might use...

    if start_date:
        if start_gran == 3:
            start_str = start_date.strftime(full_format)
        elif start_gran == 4:
            start_str = start_date.strftime(month_year_format)
        else:
            start_str = start_date.strftime(year_format)

    if end_date:
        if end_gran == 3:
            end_str = end_date.strftime(full_format)
        elif end_gran == 4:
            end_str = end_date.strftime(month_year_format)
        else:
            end_str = end_date.strftime(year_format)

    # Now make the final strings we'll return:

    if start_date and end_date:
        # A default which will be overridden in many cases. This covers:
        # 1 February 2017 to 3 March 2018
        # 1 February 2017 to March 2018
        # 1 February 2017 to 2018
        # February 2017 to 3 March 2018
        # February 2017 to March 2018
        # February 2017 to 2018
        # 2017 to 3 March 2018
        # 2017 to March 2018
        # 2017 to 2018
        output = period_format_long.format(start_str, end_str)

        if (start_gran == 4 or end_gran == 4) and same_month:
            # Only have enough to output 'February 2017'.
            output = start_str

        elif (start_gran == 6 or end_gran == 6) and same_year:
            # Only have enough to output '2017'.
            output = start_str

        elif start_gran == 3:
            if end_gran == 3:
                if same_day:
                    # 1 February 2017
                    output = start_str

                elif same_month:
                    # 1–6 February 2017
                    output = period_format_short.format(
                        start_date.strftime(day_format), end_str
                    )
                elif same_year:
                    # 1 February to 3 March 2017
                    output = period_format_long.format(
                        start_date.strftime(day_month_format), end_str
                    )
            elif end_gran == 4 and same_year:
                # 1 February to March 2017
                output = period_format_long.format(
                    start_date.strftime(day_month_format), end_str
                )
        elif start_gran == 4:
            if end_gran == 3:
                if same_year:
                    # February to 3 March 2017
                    output = period_format_long.format(
                        start_date.strftime(month_format), end_str
                    )
            elif end_gran == 4 and same_year:
                # February to March 2017
                output = period_format_long.format(
                    start_date.strftime(month_format), end_str
                )
    elif end_date:
        # Only an end_date.
        if end_gran == 3:  # noqa: SIM108
            # Finished on 1 February 2017
            output = f"Finished on {end_str}"
        else:
            # Finished in February 2017
            # Finished in 2017
            output = f"Finished in {end_str}"

    else:
        # No end_date: the reading has started, but not ended.
        if start_gran == 3:
            # Started on 1 February 2017
            output = f"Started on {start_str}"

        else:
            # Started in February 2017
            # Started in 2017
            output = f"Started in {start_str}"

    return output



from django.urls import include, path

from spectator.core.apps import spectator_apps

# The aim of this is to:
# a) Make it easy to include Spectator's URLs in a project with a single line.
# b) Also make it easy to only include parts of the project, if necessary.

app_name = "spectator"

urlpatterns = [
    path("", include("spectator.core.urls.core")),
    path("creators/", include("spectator.core.urls.creators")),
]

if spectator_apps.is_enabled("events"):
    urlpatterns.append(
        path("events/", include("spectator.events.urls", namespace="events")),
    )

if spectator_apps.is_enabled("reading"):
    urlpatterns.append(
        path("reading/", include("spectator.reading.urls", namespace="reading")),
    )



import io
import os

import piexif
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.db import models
from django.urls import reverse
from hashids import Hashids
from imagekit.models import ImageSpecField
from imagekit.processors import ResizeToFit

from . import app_settings
from .fields import NaturalSortField
from .managers import CreatorManager


class TimeStampedModelMixin(models.Model):
    "Should be mixed in to all models."

    time_created = models.DateTimeField(
        auto_now_add=True, help_text="The time this item was created in the database."
    )
    time_modified = models.DateTimeField(
        auto_now=True, help_text="The time this item was last saved to the database."
    )

    class Meta:
        abstract = True


class SluggedModelMixin(models.Model):
    """
    Adds a `slug` field which is generated from a Hashid of the model's pk.

    `slug` is generated on save, if it doesn't already exist.

    In theory we could use the Hashid'd slug in reverse to get the object's
    pk (e.g. in a view). But we're not relying on that, and simply using
    Hashid as a good method to generate unique, short URL-friendly slugs.
    """

    slug = models.SlugField(max_length=10, null=False, blank=True)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

        if not self.slug:
            self.slug = self._generate_slug(self.pk)
            # Don't want to insert again, if that's what was forced:
            kwargs["force_insert"] = False
            self.save(*args, **kwargs)

    def _generate_slug(self, value):
        """
        Generates a slug using a Hashid of `value`.
        """
        alphabet = app_settings.SLUG_ALPHABET
        salt = app_settings.SLUG_SALT

        hashids = Hashids(alphabet=alphabet, salt=salt, min_length=5)

        return hashids.encode(value)


def thumbnail_upload_path(instance, filename):
    """For ImageFields' upload_to attribute.
    e.g. '[MEDIA_ROOT]reading/publications/pok2d/my_cover_image.jpg'
    """
    # e.g. "publications" or "events":
    folder = f"{instance.__class__.__name__}s".lower()

    # This is kludgy, but...
    if folder == "publications":
        path = app_settings.READING_DIR_BASE
    elif folder == "events":
        path = app_settings.EVENTS_DIR_BASE
    else:
        msg = "No base directory set for this app's thumbnails"
        raise NotImplementedError(msg)

    return os.path.join(path, folder, instance.slug, filename)


class ThumbnailModelMixin(models.Model):
    """
    Model mixin used to add a thumbnail ImageField, and associated
    fields in different sizes, powered by django-imagekit.

    * thumbnail - The uploaded image, stored in MEDIA
    * list_thumbnail - The smallest size, for lists of objects
    * list_thumbnail_2x - Retina version of list_thumbnail
    * detail_thumbnail - Used for detail pages
    * detail_thumbnail_2x - Retina version of detail_thumbnail

    Specify the dimensions using the THUMBNAIL_LIST_SIZE and
    THUMBNAIL_DETAIL_SIZE Django settings.
    """

    thumbnail = models.ImageField(
        upload_to=thumbnail_upload_path, null=False, blank=True, default=""
    )

    # Common ImageSpecField arguments:
    thumbnail_kwargs = {
        "source": "thumbnail",
        "format": "JPEG",
        "options": {"quality": 80},
    }

    # Calculate dimensions for 2x images:
    list_thumbnail_2x_dimensions = [d * 2 for d in app_settings.THUMBNAIL_LIST_SIZE]
    detail_thumbnail_2x_dimensions = [d * 2 for d in app_settings.THUMBNAIL_DETAIL_SIZE]

    list_thumbnail = ImageSpecField(
        processors=[ResizeToFit(*app_settings.THUMBNAIL_LIST_SIZE)], **thumbnail_kwargs
    )

    list_thumbnail_2x = ImageSpecField(
        processors=[ResizeToFit(*list_thumbnail_2x_dimensions)], **thumbnail_kwargs
    )

    detail_thumbnail = ImageSpecField(
        processors=[ResizeToFit(*app_settings.THUMBNAIL_DETAIL_SIZE)],
        **thumbnail_kwargs,
    )

    detail_thumbnail_2x = ImageSpecField(
        processors=[ResizeToFit(*detail_thumbnail_2x_dimensions)], **thumbnail_kwargs
    )

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        """
        Overridden so that we can save the original thumbnail.
        So we can tell whether it's changed in save().
        """
        super().__init__(*args, **kwargs)
        self.__original_thumbnail_name = self.thumbnail.name

    def save(self, *args, **kwargs):
        """
        Move thumbnail file to correct location, and remove any location EXIF data.

        Ensure the uploaded thumbnail is in a directory for this object, with
        self.slug in the path.

        If adding a new object, e.g. in Admin's add screen, then the thumbnail
        is uploaded and saved BEFORE the object is saved. Saving the object gives
        it a pk, on which the slug is based. So we ensure the thumbnail is eventually
        saved the correct path here.
        """
        if self.pk is None and self.thumbnail:
            # The thumbnail will have been saved to a directory that doesn't have the
            # slug name, because the slug hasn't been created yet.
            # So, grab the thumbnail, save the model without it, and put the thumbnail
            # back. So when we save again, it will be put in the correct place.
            saved_thumbnail = self.thumbnail
            self.thumbnail = None
            super().save(*args, **kwargs)
            self.thumbnail = saved_thumbnail
            kwargs["force_insert"] = False

        super().save(*args, **kwargs)

        if self.thumbnail and self.__original_thumbnail_name != self.thumbnail.name:
            # New thumbnail; remove GPS data.
            self.sanitize_thumbnail_exif_data()

        # Set the original to whatever the current thumbnail is, so we
        # can tell if it changes again.
        self.__original_thumbnail_name = self.thumbnail.name

    def sanitize_thumbnail_exif_data(self):
        """
        If the thumbnail has any GPS data in its EXIF data, remove it.
        """
        if self.thumbnail:
            # We can't just do piexif.load(self.thumbnail.path) in case files are
            # stored on S3 or similar, where that doesn't work.
            # So, we use default_storage:
            file = default_storage.open(self.thumbnail.name, mode="rb")
            file_data = file.read()
            file.close()

            # Get the EXIF data from the file contents as a dict:
            exif_dict = piexif.load(file_data)

            if "GPS" in exif_dict and len(exif_dict["GPS"]) > 0:
                exif_dict["GPS"] = {}
                exif_bytes = piexif.dump(exif_dict)

                # Create a temporary file with existing data:
                temp_file = io.BytesIO(file_data)
                # Insert our updated exif data into it:
                piexif.insert(exif_bytes, file_data, temp_file)

                # Remove existing image from disk (or else, when we save the new one,
                # we'll end up with both files, the new one with a different name):
                filename = os.path.basename(self.thumbnail.name)
                self.thumbnail.delete(save=False)

                # Finally, save the temporary file as the new thumbnail:
                self.thumbnail.save(
                    filename, ContentFile(temp_file.getvalue()), save=False
                )


class BaseRole(TimeStampedModelMixin, models.Model):
    """
    Base class for linking a Creator to a Book, Event, Movie, etc.

    Child classes should add fields like:

        creator = models.ForeignKey('spectator_core.Creator', blank=False,
                    on_delete=models.CASCADE, related_name='publication_roles')

        publication = models.ForeignKey('spectator_reading.Publication',
                    on_delete=models.CASCADE, related_name='roles')
    """

    role_name = models.CharField(
        null=False,
        blank=True,
        max_length=50,
        help_text="e.g. 'Headliner', 'Support', 'Editor', 'Illustrator', "
        "'Director', etc. Optional.",
    )

    role_order = models.PositiveSmallIntegerField(
        null=False,
        blank=False,
        default=1,
        help_text="The order in which the Creators will be listed.",
    )

    class Meta:
        abstract = True
        ordering = (
            "role_order",
            "role_name",
        )

    def __str__(self):
        if self.role_name:
            return f"{self.creator} ({self.role_name})"
        else:
            return str(self.creator)


class Creator(TimeStampedModelMixin, SluggedModelMixin, models.Model):
    """
    A person or a group/company/organisation that is responsible for making all
    or part of a book, play, movie, gig, etc.

    Get the things they've worked on:

        creator = Creator.objects.get(pk=1)

        # Just Publication titles:
        for publication in creator.publications.distinct():
            print(publication.title)

        # You can do similar to that to get lists of `events`, `movies` and,
        # `plays` the Creator was involved with.


        # Or Publications including the Creator and their role:
        for role in creator.publication_roles.all():
            print(role.publication, role.creator, role.role_name)

        # Similarly for Event roles:
        for role in creator.event_roles.all():
            print(role.event, role.creator, role.role_name)

        # And for Work roles:
        for role in creator.work_roles.all():
            print(role.work, role.creator, role.role_name)
    """

    class Kind(models.TextChoices):
        INDIVIDUAL = "individual", "Individual"
        GROUP = "group", "Group"

    name = models.CharField(
        max_length=255, help_text="e.g. 'Douglas Adams' or 'The Long Blondes'."
    )

    name_sort = NaturalSortField(
        "name",
        max_length=255,
        default="",
        help_text="Best for sorting groups. e.g. 'long blondes, the' or "
        "'adams, douglas'.",
    )

    kind = models.CharField(
        max_length=20, choices=Kind.choices, default=Kind.INDIVIDUAL
    )

    objects = CreatorManager()

    class Meta:
        ordering = ("name_sort",)

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse("spectator:creators:creator_detail", kwargs={"slug": self.slug})

    @property
    def sort_as(self):
        "Used by the NaturalSortField."
        if self.kind == "individual":
            return "person"
        else:
            return "thing"

    def get_events(self):
        """
        All Events they're involved with, eliminating duplicates that occur
        with self.events.all() if they have multiple roles on the Event.
        """
        return self.events.distinct()

    def get_works(self):
        "All kinds of Work."
        return self.works.distinct()

    def get_classical_works(self):
        return self.works.filter(kind="classicalwork").distinct()

    def get_dance_pieces(self):
        return self.works.filter(kind="dancepiece").distinct()

    def get_exhibitions(self):
        return self.works.filter(kind="exhibition").distinct()

    def get_movies(self):
        return self.works.filter(kind="movie").distinct()

    def get_plays(self):
        return self.works.filter(kind="play").distinct()



from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models import Count

from .apps import spectator_apps


class CreatorManager(models.Manager):
    def by_publications(self):
        """
        The Creators who have been most-read, ordered by number of read
        publications (ignoring if any of those publicatinos have been read
        multiple times.)

        Each Creator will have a `num_publications` attribute.
        """
        if not spectator_apps.is_enabled("reading"):
            msg = (
                "To use the CreatorManager.by_publications() method, "
                "'spectator.reading' must by in INSTALLED_APPS."
            )
            raise ImproperlyConfigured(msg)

        qs = self.get_queryset()

        qs = (
            qs.exclude(publications__reading__isnull=True)
            .exclude(publications__reading__is_finished=False)
            .annotate(num_publications=Count("publications"))
            .order_by("-num_publications", "name_sort")
        )

        return qs

    def by_readings(self, role_names=None):
        """
        The Creators who have been most-read, ordered by number of readings.

        By default it will only include Creators whose role was left empty,
        or is 'Author'.

        Each Creator will have a `num_readings` attribute.
        """
        if role_names is None:
            role_names = ["", "Author"]

        if not spectator_apps.is_enabled("reading"):
            msg = (
                "To use the CreatorManager.by_readings() method, 'spectator.reading' "
                "must by in INSTALLED_APPS."
            )
            raise ImproperlyConfigured(msg)

        qs = self.get_queryset()

        qs = (
            qs.filter(publication_roles__role_name__in=role_names)
            .filter(publications__reading__is_finished=True)
            .exclude(publications__reading__isnull=True)
            .annotate(num_readings=Count("publications__reading", distinct=True))
            .order_by("-num_readings", "name_sort")
        )

        return qs

    def by_events(self, kind=None):
        """
        Get the Creators involved in the most Events.

        This only counts Creators directly involved in an Event.
        i.e. if a Creator is the director of a movie Work, and an Event was
        a viewing of that movie, that Event wouldn't count. Unless they were
        also directly involved in the Event (e.g. speaking after the movie).

        kind - If supplied, only Events with that `kind` value will be counted.
        """
        if not spectator_apps.is_enabled("events"):
            msg = (
                "To use the CreatorManager.by_events() method, 'spectator.events' "
                "must by in INSTALLED_APPS."
            )
            raise ImproperlyConfigured(msg)

        qs = self.get_queryset()

        if kind is not None:
            qs = qs.filter(events__kind=kind)

        qs = qs.annotate(num_events=Count("events", distinct=True)).order_by(
            "-num_events", "name_sort"
        )

        return qs

    def by_works(self, kind=None, role_name=None):
        """
        Get the Creators involved in the most Works.

        kind - If supplied, only Works with that `kind` value will be counted.
        role_name - If supplied, only Works on which the role is that will be counted.

        e.g. To get all 'movie' Works on which the Creators had the role 'Director':

            Creator.objects.by_works(kind='movie', role_name='Director')
        """
        if not spectator_apps.is_enabled("events"):
            msg = (
                "To use the CreatorManager.by_works() method, 'spectator.events' "
                "must by in INSTALLED_APPS."
            )
            raise ImproperlyConfigured(msg)

        qs = self.get_queryset()

        filter_kwargs = {}

        if kind is not None:
            filter_kwargs["works__kind"] = kind

        if role_name is not None:
            filter_kwargs["work_roles__role_name"] = role_name

        if filter_kwargs:
            qs = qs.filter(**filter_kwargs)

        qs = qs.annotate(num_works=Count("works", distinct=True)).order_by(
            "-num_works", "name_sort"
        )

        return qs



# Generated by Django 2.0 on 2018-01-27 16:53

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0018_remove_old_movie_play_fields"),
    ]

    operations = [
        migrations.RenameField(
            model_name="classicalworkselection",
            old_name="classical_work",
            new_name="work",
        ),
        migrations.RenameField(
            model_name="dancepieceselection",
            old_name="dance_piece",
            new_name="work",
        ),
        migrations.RenameField(
            model_name="movieselection",
            old_name="movie",
            new_name="work",
        ),
        migrations.RenameField(
            model_name="playselection",
            old_name="play",
            new_name="work",
        ),
    ]



# Generated by Django 2.0 on 2018-01-26 09:16

from django.db import migrations


class Migration(migrations.Migration):
    """
    Now we've copied their data to the new movies and plays m2m fields,
    we can remove the old movie and play fk fields.
    """

    dependencies = [
        ("spectator_events", "0017_copy_movies_and_plays_data"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="event",
            name="movie",
        ),
        migrations.RemoveField(
            model_name="event",
            name="play",
        ),
    ]



# Generated by Django 2.0 on 2018-01-26 09:07

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("spectator_events", "0015_rename_classical_dance_on_event"),
    ]

    operations = [
        migrations.CreateModel(
            name="MovieSelection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "order",
                    models.PositiveSmallIntegerField(
                        default=1, help_text="Position on the Event programme."
                    ),
                ),
            ],
            options={"verbose_name": "movie selection", "ordering": ("order",)},
        ),
        migrations.CreateModel(
            name="PlaySelection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "order",
                    models.PositiveSmallIntegerField(
                        default=1, help_text="Position on the Event programme."
                    ),
                ),
            ],
            options={"verbose_name": "play selection", "ordering": ("order",)},
        ),
        migrations.AlterField(
            model_name="event",
            name="movie",
            field=models.ForeignKey(
                blank=True,
                help_text="Only used if event is of 'Movie' kind.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="old_movie",
                to="spectator_events.Movie",
            ),
        ),
        migrations.AlterField(
            model_name="event",
            name="play",
            field=models.ForeignKey(
                blank=True,
                help_text="Only used if event is of 'Play' kind.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="old_play",
                to="spectator_events.Play",
            ),
        ),
        migrations.AddField(
            model_name="playselection",
            name="event",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="play_selections",
                to="spectator_events.Event",
            ),
        ),
        migrations.AddField(
            model_name="playselection",
            name="play",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="events",
                to="spectator_events.Play",
            ),
        ),
        migrations.AddField(
            model_name="movieselection",
            name="event",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="movie_selections",
                to="spectator_events.Event",
            ),
        ),
        migrations.AddField(
            model_name="movieselection",
            name="movie",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="events",
                to="spectator_events.Movie",
            ),
        ),
        migrations.AddField(
            model_name="event",
            name="movies",
            field=models.ManyToManyField(
                blank=True,
                help_text="Only used if event is of 'Movie' kind.",
                through="spectator_events.MovieSelection",
                to="spectator_events.Movie",
            ),
        ),
        migrations.AddField(
            model_name="event",
            name="plays",
            field=models.ManyToManyField(
                blank=True,
                help_text="Only used if event is of 'Play' kind.",
                through="spectator_events.PlaySelection",
                to="spectator_events.Play",
            ),
        ),
    ]



from django.db import models
from django.db.models import Count


class VenueManager(models.Manager):
    def by_visits(self, event_kind=None):
        """
        Gets Venues in order of how many Events have been held there.
        Adds a `num_visits` field to each one.

        event_kind filters by kind of Event, e.g. 'theatre', 'cinema', etc.
        """
        qs = self.get_queryset()

        if event_kind is not None:
            qs = qs.filter(event__kind=event_kind)

        qs = qs.annotate(num_visits=Count("event")).order_by("-num_visits", "name_sort")

        return qs


class WorkManager(models.Manager):
    def by_views(self, kind=None):
        """
        Gets Works in order of how many times they've been attached to
        Events.

        kind is the kind of Work, e.g. 'play', 'movie', etc.
        """
        qs = self.get_queryset()

        if kind is not None:
            qs = qs.filter(kind=kind)

        qs = qs.annotate(num_views=Count("event")).order_by("-num_views", "title_sort")

        return qs
